import pytest
from io import StringIO
from types import SimpleNamespace
from enum import Enum, auto

# Import the PUT (package under test)
import hdl21 as h


def test_version():
    assert h.__version__ == "0.1.1"


def test_module1():
    """ Initial Module Test """

    @h.module
    class M1:
        a = h.Input()
        b = h.Output()
        c = h.Inout()
        d = h.Port()
        e = h.Signal()
        f = h.Signal()

    assert isinstance(M1, h.Module)
    assert isinstance(M1.ports, dict)
    assert isinstance(M1.signals, dict)
    assert isinstance(M1.instances, dict)
    assert "a" in M1.ports
    assert "a" not in M1.signals
    assert "b" in M1.ports
    assert "b" not in M1.signals
    assert "c" in M1.ports
    assert "c" not in M1.signals
    assert "d" in M1.ports
    assert "d" not in M1.signals
    assert "e" in M1.signals
    assert "e" not in M1.ports
    assert "f" in M1.signals
    assert "f" not in M1.ports
    assert M1.e is not M1.f


def test_module2():
    @h.module
    class M1:
        s = h.Input()

    @h.module
    class M2:
        q = h.Signal()
        i = M1(s=q)

    assert isinstance(M1, h.Module)
    assert isinstance(M2, h.Module)
    assert isinstance(M1.ports["s"], h.Signal)
    assert isinstance(M2.signals["q"], h.Signal)
    assert isinstance(M2.instances["i"], h.Instance)

    # Test the getattr magic for signals, ports, and instances
    assert isinstance(M2.i, h.Instance)
    assert M2.instances["i"] is M2.i
    assert isinstance(M2.q, h.Signal)
    assert M2.signals["q"] is M2.q
    assert isinstance(M1.s, h.Signal)
    assert M1.ports["s"] is M1.s


def test_generator1():
    @h.paramclass
    class MyParams:
        w = h.Param(dtype=int, desc="five", default=5)

    @h.generator
    def gen1(params: MyParams) -> h.Module:
        m = h.Module()
        m.i = h.Input(width=params.w)
        return m

    m = h.elaborate(gen1(MyParams(w=3)))

    assert m.name == "gen1(MyParams(w=3))"
    assert isinstance(m.i, h.Signal)


def test_generator2():
    @h.paramclass
    class P2:
        f = h.Param(dtype=float, desc="a real number", default=1e-11)

    @h.generator
    def g2(params: P2, ctx: h.Context) -> h.Module:
        # Generator which takes a Context-argument
        assert isinstance(params, P2)
        assert isinstance(ctx, h.Context)
        return h.Module()

    m = h.elaborate(g2(P2()))

    assert isinstance(m, h.Module)
    assert m.name == "g2(P2(f=1e-11))"


def test_generator3():
    @h.paramclass
    class P3:
        width = h.Param(dtype=int, desc="bit-width, maybe", default=1)

    @h.generator
    def g3a(params: P3) -> h.Module:
        return h.Module()

    @h.generator
    def g3b(params: P3, ctx: h.Context) -> h.Module:
        return h.Module()

    M = h.Module(name="M")
    p3a = P3()
    p3b = P3(width=5)

    @h.module
    class HasGen:
        a = g3a(p3a)()
        b = g3b(p3b)()
        c = M()

    # Elaborate the top module
    h.elaborate(HasGen)

    # Post-elab checks
    assert isinstance(HasGen.a, h.Instance)
    assert isinstance(HasGen.a.of, h.GeneratorCall)
    assert HasGen.a.of.gen is g3a
    assert HasGen.a.of.arg == P3()
    assert isinstance(HasGen.a.of.result, h.Module)
    assert HasGen.a.of.result.name == "g3a(P3(width=1))"
    assert isinstance(HasGen.b, h.Instance)
    assert isinstance(HasGen.b.of, h.GeneratorCall)
    assert isinstance(HasGen.b.of.result, h.Module)
    assert HasGen.b.of.result.name == "g3b(P3(width=5))"
    assert HasGen.b.of.gen is g3b
    assert HasGen.b.of.arg == P3(width=5)
    assert isinstance(HasGen.c, h.Instance)
    assert isinstance(HasGen.c.of, h.Module)
    assert HasGen.c.of is M


def test_params1():
    # Initial param-class test

    @h.paramclass
    class MyParams:
        a = h.Param(dtype=int, default=5, desc="your fave")

    assert h.isparamclass(MyParams)

    m = MyParams()
    assert isinstance(m.a, int)
    assert m.a == 5
    assert isinstance(m.__params__["a"], h.Param)
    assert m.defaults() == dict(a=5)
    assert m.descriptions() == dict(a="your fave")


def test_params2():
    @h.paramclass
    class Pc2:
        # Required
        r = h.Param(dtype=str, desc="required")
        # Optional
        o = h.Param(dtype=str, default="hmmm", desc="optional")

    assert h.isparamclass(Pc2)
    assert Pc2.defaults() == dict(o="hmmm")
    assert Pc2.descriptions() == dict(r="required", o="optional")

    p = Pc2(r="provided")
    assert isinstance(p.r, str)
    assert p.r == "provided"
    assert isinstance(p.o, str)
    assert p.o == "hmmm"
    assert p.defaults() == dict(o="hmmm")
    assert p.descriptions() == dict(r="required", o="optional")


def test_params3():
    # Test parameters with a creation-time list-tuple conversion
    @h.paramclass
    class HasTuple:
        t = h.Param(dtype=tuple, desc="Go ahead, try a list")

    ht = HasTuple(t=[1, 2, 3])
    assert isinstance(ht.t, tuple)
    assert ht.t == (1, 2, 3)


def test_params4():
    # Test some param-class nesting
    from dataclasses import asdict

    @h.paramclass
    class Inner:
        i = h.Param(dtype=int, desc="Inner int-field")

    @h.paramclass
    class Outer:
        inner = h.Param(dtype=Inner, desc="Inner fields")
        f = h.Param(dtype=float, desc="A float", default=3.14159)

    o = Outer(inner=Inner(11))
    assert isinstance(o, Outer)
    assert isinstance(o.inner, Inner)
    assert o.inner == Inner(11)
    assert o.inner.i == 11
    assert o.f == 3.14159

    # Create from a (nested) dictionary
    d1 = {"inner": {"i": 11}, "f": 22.2}
    o1 = Outer(**d1)
    uname1 = h.params._unique_name(o1)
    # Convert back to another dictionary
    d2 = asdict(o1)
    # And check they line up
    assert d1 == d2
    # Round-trip back to an `Outer`
    o2 = Outer(**d2)
    uname2 = h.params._unique_name(o2)
    assert uname1 == uname2
    assert uname1 == "Outer(3dcc309796996b3a8a61db66631c5a93)"


def test_bad_params1():
    # Test a handful of errors Params and paramclasses should raise.

    from hdl21 import (
        paramclass,
        Param,
        ValidationError,
        FrozenInstanceError,
    )

    with pytest.raises(RuntimeError):
        # Test that creating a paramclass with parent-class(es) fails

        @paramclass
        class C(TabError):  # Of course this is a sub-class of the best built-in class
            ...

    @paramclass
    class C:
        a = Param(dtype=int, desc="Gonna Fail!")

    with pytest.raises(RuntimeError):
        # Test that sub-classing a paramclass fails

        class D(C):
            ...

    with pytest.raises(TypeError):
        # Test that missing arguments fail
        c = C()

    with pytest.raises(ValidationError):
        # Test invalid argument types fail

        c = C(a=TabError)

    with pytest.raises(FrozenInstanceError):
        # Test that attempts at mutating paramclasses fail
        c = C(a=3)
        c.a = 4

    with pytest.raises(RuntimeError):
        # Test "no Module sub-classing"

        class E(h.Module):
            ...

    with pytest.raises(RuntimeError):
        # Test "no decorating inherited types"

        @h.module
        class E2(TabError):
            ...

    with pytest.raises(RuntimeError):
        # Test bad parameter names

        @h.paramclass
        class P:
            descriptions = h.Param(dtype=str, desc="", default="BAD!")

    with pytest.raises(RuntimeError):
        # Test bad parameter names

        @h.paramclass
        class P:
            defaults = h.Param(dtype=str, desc="", default="BAD!")

    with pytest.raises(RuntimeError):
        # Test non-params in `@paramclass`

        @h.paramclass
        class P:
            something = 11

    with pytest.raises(RuntimeError):
        # Test a bad argument type
        h.params._unique_name(33)


def test_array1():
    @h.module
    class InArray:
        inp = h.Input()
        out = h.Output()

    m = h.Module(name="HasArray")
    m.s1 = h.Signal(width=8)
    m.s2 = h.Signal(width=1)
    m.arr = h.InstArray(InArray, 8)
    m.arr.inp = m.s1
    m.arr.out = m.s2

    assert m.name == "HasArray"


def test_array2():
    """ Basic Instance-Array Test """
    a = h.Module(name="a")
    a.inp = h.Port(width=1)
    a.out = h.Port(width=1)

    @h.module
    class HasArray2:
        s1 = h.Signal(width=8)
        s2 = h.Signal(width=1)
        arr = h.InstArray(a, 8)(inp=s1, out=s2)

    assert len(HasArray2.instances) == 0
    assert len(HasArray2.instarrays) == 1

    # Elaborate, flattening arrays along the way
    h.elaborate(HasArray2)

    # Post-elab checks
    assert len(HasArray2.instances) == 8
    assert len(HasArray2.instarrays) == 0


def test_cycle1():
    """ Test cyclical connection-graphs, i.e. a back-to-back pair of instances """

    @h.module
    class Thing:
        inp = h.Input()
        out = h.Output()

    @h.module
    class BackToBack:
        t1 = Thing()
        t2 = Thing(inp=t1.out, out=t1.inp)

    b = h.elaborate(BackToBack)

    assert isinstance(b.t1.inp, h.Signal)
    assert isinstance(b.t1.out, h.Signal)
    assert isinstance(b.t2.inp, h.Signal)
    assert isinstance(b.t2.out, h.Signal)

    # Doing the same thing in procedural code
    b2 = h.Module(name="BackToBack2")
    b2.t1 = Thing()
    b2.t2 = Thing()
    b2.t2.inp = b2.t1.out
    b2.t2.out = b2.t1.inp

    assert isinstance(b2.t1.inp, h.PortRef)
    assert isinstance(b2.t1.out, h.PortRef)
    assert isinstance(b2.t2.inp, h.PortRef)
    assert isinstance(b2.t2.out, h.PortRef)

    b2 = h.elaborate(b2)

    assert len(b2.instances) == 2
    assert len(b2.instarrays) == 0
    assert len(b2.interfaces) == 0
    assert len(b2.ports) == 0
    assert len(b2.signals) == 2
    assert "_t2_inp_t1_out_" in b2.signals
    assert "_t2_out_t1_inp_" in b2.signals


def test_gen3():
    @h.paramclass
    class MyParams:
        w = h.Param(dtype=int, desc="Input bit-width. Required")

    @h.generator
    def MyThirdGenerator(params: MyParams) -> h.Module:
        @h.module
        class Inner:
            i = h.Input(width=params.w)
            o = h.Output(width=2 * params.w)

        # Instantiate that in another Module
        @h.module
        class Outer:
            i = h.Signal(width=params.w)
            o = h.Signal(width=2 * params.w)
            inner = Inner(i=i, o=o)

        # And manipulate that some more too
        Outer.inp = h.Input(width=params.w)
        return Outer

    h.elaborate(MyThirdGenerator(MyParams(1)))

    # FIXME: post-elab checks


def test_prim1():
    # First test of transistor primitives
    params = h.Mos.Params()
    pmos = h.Pmos(params)
    nmos = h.Nmos(params)

    @h.module
    class HasMos:
        # Two transistors wired in parallel
        p = pmos()
        n = nmos(g=p.g, d=p.d, s=p.s, b=p.b)

    h.elaborate(HasMos)
    # FIXME: post-elab checks


def test_prim2():
    @h.module
    class HasPrims:
        p = h.Signal()
        n = h.Signal()

        _rp = h.R.Params(r=50)
        r = h.Resistor(_rp)(p=p, n=n)

        c = h.Capacitor(h.C.Params(c=1e-12))(p=p, n=n)
        l = h.Inductor(h.L.Params(l=1e-15))(p=p, n=n)
        d = h.Diode(h.D.Params())(p=p, n=n)
        s = h.Short(h.Short.Params())(p=p, n=n)

    h.elaborate(HasPrims)
    # FIXME: post-elab checks


def test_prim_proto1():
    # Test round-tripping primitives through proto

    @h.module
    class HasPrims:
        p = h.Signal()
        n = h.Signal()

        # Wire up a bunch of two-terminal primitives in parallel
        _rp = h.R.Params(r=50)
        r = h.Resistor(_rp)(p=p, n=n)
        _cp = h.C.Params(c=1e-12)
        c = h.Capacitor(_cp)(p=p, n=n)
        _lp = h.L.Params(l=1e-15)
        l = h.Inductor(_lp)(p=p, n=n)
        _dp = h.D.Params()
        d = h.Diode(_dp)(p=p, n=n)
        _sp = h.Short.Params()
        s = h.Short(_sp)(p=p, n=n)

    ppkg = h.to_proto(HasPrims)

    assert isinstance(ppkg, h.proto.Package)
    assert len(ppkg.modules) == 1
    assert ppkg.domain == ""

    # Check the proto-Module
    pm = ppkg.modules[0]
    assert isinstance(pm, h.proto.Module)
    assert pm.name == "hdl21.tests.test_hdl21.HasPrims"
    assert len(pm.ports) == 0
    assert len(pm.signals) == 2
    assert len(pm.instances) == 5
    assert len(pm.default_parameters) == 0
    for inst in pm.instances:
        assert isinstance(inst, h.proto.Instance)
        assert inst.module.WhichOneof("to") == "external"
        assert inst.module.external.domain in ["hdl21.ideal", "hdl21.primitives"]

    ns = h.from_proto(ppkg)

    assert isinstance(ns, SimpleNamespace)
    ns = ns.hdl21.tests.test_hdl21
    assert isinstance(ns, SimpleNamespace)
    assert hasattr(ns, "HasPrims")
    HasPrims = ns.HasPrims
    assert isinstance(HasPrims, h.Module)
    assert len(HasPrims.ports) == 0
    assert len(HasPrims.signals) == 2
    assert len(HasPrims.instances) == 5
    for inst in HasPrims.instances.values():
        assert isinstance(inst._resolved, h.primitives.PrimitiveCall)


def test_intf1():
    # Create an interface

    i1 = h.Interface(name="MyFirstInterface")
    i1.s1 = h.Signal()
    i1.s2 = h.Signal()

    assert isinstance(i1, h.Interface)
    assert isinstance(i1.s1, h.Signal)
    assert isinstance(i1.s2, h.Signal)

    ii1 = i1()
    assert isinstance(ii1, h.InterfaceInstance)
    assert ii1.role is None
    assert ii1.port == False


def test_intf2():
    # Wire up a few Modules via interfaces

    MySecondInterface = h.Interface(name="MySecondInterface")
    MySecondInterface.s = h.Signal()

    m1 = h.Module(name="M1")
    m1.i = MySecondInterface(port=True)
    m2 = h.Module(name="M2")
    m2.i = MySecondInterface(port=True)

    # Now create a parent Module connecting the two
    m3 = h.Module(name="M3")
    m3.i1 = m1()
    m3.i2 = m2(i=m3.i1.i)
    assert "_i2_i_i1_i_" not in m3.namespace

    # First run the "implicit interfaces" pass, and see that an explicit one is created
    from hdl21.elab import ElabPass

    m3 = h.elaborate(m3, passes=[ElabPass.IMPLICIT_INTERFACES])
    assert isinstance(m3, h.Module)
    assert isinstance(m3.i1, h.Instance)
    assert isinstance(m3.i2, h.Instance)
    assert "_i2_i_i1_i_" in m3.namespace

    # Now elaborate it the rest of the way, to scalar signals
    m3 = h.elaborate(m3)
    assert "_i2_i_i1_i_" not in m3.namespace
    assert "__i2_i_i1_i__s_" in m3.namespace
    assert isinstance(m3.get("__i2_i_i1_i__s_"), h.Signal)
    assert m3.get("__i2_i_i1_i__s_") in m3.i1.conns.values()


def test_intf3():
    # Test the interface-definition decorator

    @h.interface
    class Diff:  # Differential Signal Interface
        p = h.Signal()
        n = h.Signal()

    @h.interface
    class DisplayPort:  # DisplayPort, kinda
        main_link = Diff()
        aux = h.Signal()

    assert isinstance(DisplayPort, h.Interface)
    assert isinstance(DisplayPort(), h.InterfaceInstance)
    assert isinstance(DisplayPort.main_link, h.InterfaceInstance)
    assert isinstance(DisplayPort.main_link.p, h.PortRef)
    assert isinstance(DisplayPort.main_link.n, h.PortRef)
    assert isinstance(DisplayPort.aux, h.Signal)
    assert isinstance(Diff, h.Interface)
    assert isinstance(Diff(), h.InterfaceInstance)
    assert isinstance(Diff.p, h.Signal)
    assert isinstance(Diff.n, h.Signal)


def test_intf4():
    # Test interface roles

    from enum import Enum, EnumMeta, auto

    @h.interface
    class Diff:  # Differential Signal Interface
        p = h.Signal()
        n = h.Signal()

    @h.interface
    class HasRoles:  # An Interface with Roles
        class Roles(Enum):  # USB-Style Role Nomenclature
            HOST = auto()
            DEVICE = auto()

        # Create signals going in either direction
        tx = h.Signal(src=Roles.HOST, dest=Roles.DEVICE)
        rx = h.Signal(src=Roles.DEVICE, dest=Roles.HOST)

        # And create differential versions thereof
        txd = Diff(src=Roles.HOST, dest=Roles.DEVICE)
        rxd = Diff(src=Roles.DEVICE, dest=Roles.HOST)

    hr = HasRoles()
    assert isinstance(HasRoles, h.Interface)
    assert isinstance(HasRoles.roles, EnumMeta)
    assert isinstance(HasRoles.Roles, EnumMeta)
    assert isinstance(hr, h.InterfaceInstance)
    assert isinstance(HasRoles.tx, h.Signal)
    assert isinstance(HasRoles.rx, h.Signal)
    assert isinstance(HasRoles.txd, h.InterfaceInstance)
    assert isinstance(HasRoles.rxd, h.InterfaceInstance)

    @h.module
    class Host:
        # A thing with a HOST-roled interface-port
        port_ = HasRoles(port=True, role=HasRoles.Roles.HOST)

    @h.module
    class Device:
        # A thing with a DEVICE-roled interface-port
        port_ = HasRoles(port=True, role=HasRoles.Roles.DEVICE)

    @h.module
    class System:
        # Parent system-module including a host and device
        host = Host()
        dev_ = Device(port_=host.port_)

    assert isinstance(System, h.Module)
    assert isinstance(System.host, h.Instance)
    assert isinstance(System.dev_, h.Instance)
    assert "_dev__port__host_port__" not in System.namespace

    # First run the "implicit interfaces" pass, and see that an explicit one is created
    from hdl21.elab import ElabPass

    sys = h.elaborate(System, passes=[ElabPass.IMPLICIT_INTERFACES])
    assert "_dev__port__host_port__" in sys.namespace

    # Now expand the rest of the way, down to scalar signals
    # Check that interface went away, and its constituent signals replaced it
    sys = h.elaborate(sys)
    assert "_dev__port__host_port__" not in sys.namespace
    assert "__dev__port__host_port___tx_" in sys.namespace
    assert "__dev__port__host_port___rx_" in sys.namespace
    assert "__dev__port__host_port____txd_p__" in sys.namespace
    assert "__dev__port__host_port____txd_n__" in sys.namespace
    assert "__dev__port__host_port____rxd_p__" in sys.namespace
    assert "__dev__port__host_port____rxd_n__" in sys.namespace


def test_proto1():
    # First Proto-export test

    m = h.Module(name="TestProto1")
    ppkg = h.to_proto(m)
    assert isinstance(ppkg, h.proto.Package)
    assert len(ppkg.modules) == 1
    assert ppkg.domain == ""
    pm = ppkg.modules[0]
    assert isinstance(pm, h.proto.Module)
    assert pm.name == "hdl21.tests.test_hdl21.TestProto1"
    assert len(pm.ports) == 0
    assert len(pm.instances) == 0
    assert len(pm.default_parameters) == 0


def test_proto2():
    # Proto-export test with some hierarchy

    Child1 = h.Module(name="Child1")
    Child1.inp = h.Input(width=8)
    Child1.out = h.Output()
    Child2 = h.Module(name="Child2")
    Child2.inp = h.Input()
    Child2.out = h.Output(width=8)
    TestProto2 = h.Module(name="TestProto2")
    TestProto2.c1 = Child1()
    TestProto2.c2 = Child2()
    TestProto2.c2(inp=TestProto2.c1.out, out=TestProto2.c1.inp)

    ppkg = h.to_proto(TestProto2)

    assert isinstance(ppkg, h.proto.Package)
    assert len(ppkg.modules) == 3
    assert ppkg.domain == ""

    # Check the first Module in, Child1
    pm = ppkg.modules[0]
    assert isinstance(pm, h.proto.Module)
    assert pm.name == "hdl21.tests.test_hdl21.Child1"
    assert len(pm.ports) == 2
    assert len(pm.signals) == 0
    assert len(pm.instances) == 0
    assert len(pm.default_parameters) == 0

    # Check the second Module in, Child2
    pm = ppkg.modules[1]
    assert isinstance(pm, h.proto.Module)
    assert pm.name == "hdl21.tests.test_hdl21.Child2"
    assert len(pm.ports) == 2
    assert len(pm.signals) == 0
    assert len(pm.instances) == 0
    assert len(pm.default_parameters) == 0

    # And check the parent module
    pm = ppkg.modules[2]
    assert isinstance(pm, h.proto.Module)
    assert pm.name == "hdl21.tests.test_hdl21.TestProto2"
    assert len(pm.ports) == 0
    assert len(pm.instances) == 2
    assert len(pm.default_parameters) == 0


def test_proto3():
    # Proto-export test with some slicing and concatenation

    M1 = h.Module(name="M1")
    M1.p8 = h.Inout(width=8)
    M1.p1 = h.Inout(width=1)

    M2 = h.Module(name="M2")
    M2.s = h.Signal(width=4)
    M2.i = M1()
    M2.i(p1=M2.s[0])
    M2.i(p8=h.Concat(M2.s, h.Concat(M2.s[0], M2.s[1]), M2.s[2:3]))

    ppkg = h.to_proto(M2, domain="test_proto3")

    assert isinstance(ppkg, h.proto.Package)
    assert len(ppkg.modules) == 2
    assert ppkg.domain == "test_proto3"

    # Check the child module
    pm = ppkg.modules[0]
    assert isinstance(pm, h.proto.Module)
    assert pm.name == "hdl21.tests.test_hdl21.M1"
    assert len(pm.ports) == 2
    assert len(pm.signals) == 0
    assert len(pm.instances) == 0
    assert len(pm.default_parameters) == 0

    # And check the parent module
    pm = ppkg.modules[1]
    assert isinstance(pm, h.proto.Module)
    assert pm.name == "hdl21.tests.test_hdl21.M2"
    assert len(pm.ports) == 0
    assert len(pm.signals) == 1
    assert len(pm.instances) == 1
    assert len(pm.default_parameters) == 0

    ns = h.from_proto(ppkg)

    assert isinstance(ns, SimpleNamespace)
    ns = ns.hdl21.tests.test_hdl21
    assert isinstance(ns, SimpleNamespace)
    assert isinstance(ns.M1, h.Module)
    assert len(ns.M1.ports) == 2
    assert len(ns.M1.signals) == 0
    assert len(ns.M1.instances) == 0

    assert isinstance(M2, h.Module)
    assert len(M2.ports) == 0
    assert len(M2.signals) == 1
    assert len(M2.instances) == 1
    assert "s" in M2.signals
    assert "i" in M2.instances
    inst = M2.i
    assert isinstance(inst.conns["p1"], h.signal.Slice)
    assert inst.conns["p1"].signal is M2.s
    assert inst.conns["p1"].top == 0
    assert inst.conns["p1"].bot == 0
    assert isinstance(inst.conns["p8"], h.signal.Concat)
    assert len(inst.conns["p8"].parts) == 4
    assert inst.conns["p8"].parts[0] is M2.s
    assert isinstance(inst.conns["p8"].parts[0], h.signal.Signal)
    assert isinstance(inst.conns["p8"].parts[1], h.signal.Slice)
    assert isinstance(inst.conns["p8"].parts[2], h.signal.Slice)
    assert isinstance(inst.conns["p8"].parts[3], h.signal.Slice)


def test_proto_roundtrip():
    # Test protobuf round-tripping

    # Create an empty (named) Module
    M1 = h.Module(name="M1")

    # Protobuf round-trip it
    ppkg = h.to_proto(M1)
    ns = h.from_proto(ppkg)

    assert isinstance(ns, SimpleNamespace)
    ns = ns.hdl21.tests.test_hdl21
    assert isinstance(ns, SimpleNamespace)
    assert isinstance(ns.M1, h.Module)
    assert len(ns.M1.signals) == 0
    assert len(ns.M1.ports) == 0
    assert len(ns.M1.instances) == 0


def test_proto_roundtrip2():

    # Create a child/leaf Module
    M1 = h.Module(name="M1")
    M1.i = h.Input()
    M1.o = h.Output()
    M1.p = h.Port()

    # Create an instantiating parent-module
    M2 = h.Module(name="M2")
    M2.i = h.Input()
    M2.o = h.Output()
    M2.p = h.Port()
    M2.s = h.Signal()

    # Add a few instances of it
    M2.i0 = M1()
    M2.i1 = M1()
    M2.i2 = M1()
    M2.i0(i=M2.i, o=M2.i1.i, p=M2.p)
    M2.i1(i=M2.i0.o, o=M2.s, p=M2.p)
    M2.i2(i=M2.s, o=M2.o, p=M2.p)

    # Protobuf round-trip it
    ppkg = h.to_proto(M2)
    ns = h.from_proto(ppkg)

    # And check all kinda stuff about what comes back
    assert isinstance(ns, SimpleNamespace)
    M1 = ns.hdl21.tests.test_hdl21.M1
    M2 = ns.hdl21.tests.test_hdl21.M2
    assert isinstance(M1, h.Module)
    assert isinstance(M2, h.Module)

    assert len(M1.signals) == 0
    assert len(M1.ports) == 3
    assert "i" in M1.ports
    assert M1.i.direction == h.signal.PortDir.INPUT
    assert M1.i.vis == h.signal.Visibility.PORT
    assert "o" in M1.ports
    assert M1.o.direction == h.signal.PortDir.OUTPUT
    assert M1.o.vis == h.signal.Visibility.PORT
    assert "p" in M1.ports
    assert M1.p.direction == h.signal.PortDir.NONE
    assert M1.p.vis == h.signal.Visibility.PORT
    assert len(M1.instances) == 0

    assert len(M2.instances) == 3
    for i in M2.instances.values():
        assert i.of is M1
        assert i.conns["p"] is M2.p
    assert len(M2.ports) == 3
    assert len(M2.signals) == 2
    assert "s" in M2.signals
    assert "_i1_i_i0_o_" in M2.signals


def test_bigger_interfaces():
    # Test a slightly more elaborate Interface-based system

    class MsRoles(Enum):
        MS = auto()
        SL = auto()

    @h.interface
    class Jtag:
        roles = MsRoles
        tck = h.Signal(src=roles.MS, dest=roles.SL)
        tdi = h.Signal(src=roles.MS, dest=roles.SL)
        tms = h.Signal(src=roles.MS, dest=roles.SL)
        tdo = h.Signal(src=roles.SL, dest=roles.MS)

    @h.interface
    class Spi:
        roles = MsRoles
        sck = h.Signal(src=roles.MS, dest=roles.SL)
        cs = h.Signal(src=roles.MS, dest=roles.SL)
        dq = h.Signal(src=roles.SL, dest=roles.MS, width=4)

    @h.module
    class Chip:
        spi = Spi(role=MsRoles.MS, port=True)
        jtag = Jtag(role=MsRoles.SL, port=True)
        ...  # Actual internal content, which likely connects these down *many* levels of hierarchy

    @h.module
    class SpiFlash:
        # A typical flash memory with a SPI port
        spi = Spi(role=MsRoles.SL, port=True)

    @h.module
    class Board:
        # A typical embedded board, featuring a custom chip, SPI-connected flash, and JTAG port
        jtag = Jtag(role=MsRoles.SL, port=True)
        chip = Chip(jtag=jtag)
        flash = SpiFlash(spi=chip.spi)

    @h.module
    class Tester:
        # A typical test-widget with a JTAG port
        jtag = Jtag(role=MsRoles.MS, port=True)

    @h.module
    class TestSystem:
        # A system in which `Tester` can test `Board`
        jtag = Jtag()
        tester = Tester(jtag=jtag)
        board = Board(jtag=jtag)

    # Run this through elaboration
    h.elaborate(TestSystem)

    # Post-elab checks
    assert not hasattr(TestSystem, "jtag")
    assert len(TestSystem.ports) == 0
    assert len(TestSystem.signals) == 4
    assert len(TestSystem.instances) == 2
    assert TestSystem.tester.of is Tester
    assert TestSystem.board.of is Board
    assert isinstance(TestSystem.get("_jtag_tck_"), h.Signal)
    assert isinstance(TestSystem.get("_jtag_tdi_"), h.Signal)
    assert isinstance(TestSystem.get("_jtag_tdo_"), h.Signal)
    assert isinstance(TestSystem.get("_jtag_tms_"), h.Signal)
    assert isinstance(Tester.get("_jtag_tck_"), h.Signal)
    assert len(Tester.ports) == 4
    assert len(Tester.signals) == 0
    assert len(Tester.instances) == 0
    assert Tester.get("_jtag_tck_").vis == h.signal.Visibility.PORT
    assert Tester.get("_jtag_tdo_").vis == h.signal.Visibility.PORT
    assert Tester.get("_jtag_tdi_").vis == h.signal.Visibility.PORT
    assert Tester.get("_jtag_tms_").vis == h.signal.Visibility.PORT
    assert len(Board.ports) == 4
    assert len(Board.signals) == 3  # SPI signals
    assert len(Board.instances) == 2
    assert Board.chip.of is Chip
    assert Board.flash.of is SpiFlash
    assert Board.get("_jtag_tck_").vis == h.signal.Visibility.PORT
    assert Board.get("_jtag_tdo_").vis == h.signal.Visibility.PORT
    assert Board.get("_jtag_tdi_").vis == h.signal.Visibility.PORT
    assert Board.get("_jtag_tms_").vis == h.signal.Visibility.PORT
    assert Board.get("__flash_spi_chip_spi__sck_").vis == h.signal.Visibility.INTERNAL
    assert Board.get("__flash_spi_chip_spi__cs_").vis == h.signal.Visibility.INTERNAL
    assert Board.get("__flash_spi_chip_spi__dq_").vis == h.signal.Visibility.INTERNAL
    assert len(Chip.ports) == 7
    assert len(Chip.signals) == 0
    assert len(Chip.instances) == 0
    assert Chip.get("_jtag_tck_").vis == h.signal.Visibility.PORT
    assert Chip.get("_jtag_tdo_").vis == h.signal.Visibility.PORT
    assert Chip.get("_jtag_tdi_").vis == h.signal.Visibility.PORT
    assert Chip.get("_jtag_tms_").vis == h.signal.Visibility.PORT
    assert Chip.get("_spi_sck_").vis == h.signal.Visibility.PORT
    assert Chip.get("_spi_cs_").vis == h.signal.Visibility.PORT
    assert Chip.get("_spi_dq_").vis == h.signal.Visibility.PORT
    assert len(SpiFlash.ports) == 3
    assert len(SpiFlash.signals) == 0
    assert len(SpiFlash.instances) == 0
    assert SpiFlash.get("_spi_sck_").vis == h.signal.Visibility.PORT
    assert SpiFlash.get("_spi_cs_").vis == h.signal.Visibility.PORT
    assert SpiFlash.get("_spi_dq_").vis == h.signal.Visibility.PORT


def test_signal_slice1():
    # Initial test of signal slicing

    sig = h.Signal(width=10)
    sl = sig[0]
    assert isinstance(sl, h.signal.Slice)
    assert sl.top == 0
    assert sl.bot == 0
    assert sl.width == 1
    assert sl.step == 1
    assert sl.signal is sig

    sl = sig[0:4]
    assert isinstance(sl, h.signal.Slice)
    assert sl.top == 4
    assert sl.bot == 0
    assert sl.width == 5
    assert sl.signal is sig

    sl = sig[:]
    assert isinstance(sl, h.signal.Slice)
    assert sl.top == 9
    assert sl.bot == 0
    assert sl.width == 10
    assert sl.signal is sig


def test_signal_slice2():
    # Test slicing advanced features
    sl = h.Signal(width=11)[-1]
    assert sl.top == 10
    assert sl.bot == 10
    assert sl.width == 1

    sl = h.Signal(width=11)[:-1]
    assert sl.top == 9
    assert sl.bot == 0
    assert sl.width == 10

    sl = h.Signal(width=11)[-2:]
    assert sl.top == 10
    assert sl.bot == 9
    assert sl.width == 2


def test_bad_slice1():
    # Test slicing error-cases
    with pytest.raises(TypeError):
        h.Signal(width=11)[None]

    with pytest.raises(ValueError):
        h.Signal(width=11)[11]

    h.Signal(width=11)[2:1:-1]  # OK
    with pytest.raises(ValueError):
        h.Signal(width=11)[2:1:1]  # Not OK

    h.Signal(width=11)[1:9]  # OK
    with pytest.raises(ValueError):
        h.Signal(width=11)[9:1]  # Not OK


def test_signal_concat1():
    # Initial Signal concatenation test

    c = h.Concat(h.Signal(width=1), h.Signal(width=2), h.Signal(width=3))
    assert isinstance(c, h.signal.Concat)
    assert len(c.parts) == 3
    for p in c.parts:
        assert isinstance(p, h.signal.Signal)
    assert c.width == 6

    c = h.Concat(h.Signal(width=1)[0], h.Signal(width=2)[0], h.Signal(width=3)[0])
    assert isinstance(c, h.signal.Concat)
    assert len(c.parts) == 3
    for p in c.parts:
        assert isinstance(p, h.signal.Slice)
    assert c.width == 3

    c = h.Concat(
        h.Concat(h.Signal(), h.Signal()), h.Signal(width=2), h.Signal(width=2)[:]
    )
    assert isinstance(c, h.signal.Concat)
    assert len(c.parts) == 3
    assert c.width == 6


def test_slice_module1():
    # Make use of slicing and concatenation in a module

    C = h.Module(name="C")
    C.p1 = h.Port(width=1)
    C.p4 = h.Port(width=4)
    C.p7 = h.Port(width=7)

    P = h.Module(name="P")
    C.s4 = h.Signal(width=4)
    C.s2 = h.Signal(width=2)
    P.ic = C(p1=C.s4[0], p4=C.s4, p7=h.Concat(C.s4, C.s2, C.s2[0]))

    h.elaborate(P)


def test_bad_proto_naming():
    # Test some cases which generate serialization naming conflicts
    # Same Module-names in same Python module

    with pytest.raises(RuntimeError):
        # Create a naming conflict between Module definitions
        Ma1 = h.Module(name="Ma")
        Ma2 = h.Module(name="Ma")

        @h.module
        class MaParent:
            ma1 = Ma1()
            ma2 = Ma2()

        h.to_proto(MaParent)

    with pytest.raises(RuntimeError):
        # Do the same via some functions that should be generators
        def m1():
            return h.Module(name="Ma")

        def m2():
            return h.Module(name="Ma")

        @h.module
        class MaParent:
            ma1 = m1()()
            ma2 = m2()()

        h.to_proto(MaParent)


def test_generator_recall():
    """ Test multi-calling generators """

    @h.generator
    def CallMeTwice(_: h.HasNoParams) -> h.Module:
        return h.Module()

    @h.module
    class Caller:
        m1 = CallMeTwice(h.NoParams)()
        m2 = CallMeTwice(h.NoParams)()

    # Convert to proto
    ppkg = h.to_proto(Caller)
    assert len(ppkg.modules) == 2
    assert ppkg.modules[0].name == "hdl21.tests.test_hdl21.CallMeTwice(NoParams())"
    assert ppkg.modules[1].name == "hdl21.tests.test_hdl21.Caller"

    # Convert the proto-package to a netlist, and run some (very basic) checks
    nl = StringIO()
    h.netlist(ppkg, nl)
    nl = nl.getvalue()
    assert "CallMeTwice_NoParams___" in nl
    assert "Caller" in nl

    # Round-trip back from Proto to Modules
    rt = h.from_proto(ppkg)
    ns = rt.hdl21.tests.test_hdl21
    assert isinstance(ns.Caller, h.Module)
    assert isinstance(getattr(ns, "CallMeTwice(NoParams())"), h.Module)


def test_module_as_param():
    """ Test using a `Module` as a parameter-value """

    @h.paramclass
    class HasModuleParam:
        m = h.Param(dtype=h.Module, desc="A `Module` provided as a parameter")

    @h.generator
    def UsesModuleParam(params: HasModuleParam) -> h.Module:
        return params.m  # Returns the Module unmodified

    Empty = h.Module(name="Empty")
    p = HasModuleParam(m=Empty)
    m = UsesModuleParam(p)
    m = h.elaborate(m)
    assert m == Empty


def test_mos_generator():
    """ Initial test of built-in series-Mos generator """

    Mos = h.generators.Mos
    m = Mos(Mos.Params(nser=2))

    assert isinstance(m, h.GeneratorCall)
    assert isinstance(m.arg, Mos.Params)

    m = h.elaborate(m)

    assert isinstance(m, h.Module)
    assert len(m.ports) == 4
    assert m.ports == h.primitives.Mos.ports
    assert len(m.instances) == 2
    assert "unit0" in m.instances
    assert "unit1" in m.instances
    assert isinstance(m.instances["unit0"].of, h.PrimitiveCall)
    assert isinstance(m.instances["unit1"].of, h.PrimitiveCall)
    assert m.instances["unit0"].of.params == h.primitives.MosParams()
    assert m.instances["unit1"].of.params == h.primitives.MosParams()
    assert len(m.signals) == 1
    assert "_unit1_s_unit0_d_" in m.signals

    ppkg = h.to_proto(m)
    assert isinstance(ppkg, h.proto.Package)


def test_series_parallel_generator():
    """ Initial test of the general-purpose series-parallel generator """

    from hdl21.generators import SeriesPar

    @h.module
    class M:  # Unit cell
        a, b, c, d, e, f, g = h.Ports(7)

    params = SeriesPar.Params(unit=M, npar=2, nser=2, series_conns=["a", "b"])
    m = SeriesPar(params)

    assert isinstance(m, h.GeneratorCall)
    assert isinstance(m.arg, SeriesPar.Params)

    m = h.elaborate(m)

    assert isinstance(m, h.Module)
    assert len(m.ports) == 7
    assert m.ports == M.ports
    assert len(m.instances) == 4
    assert "unit_0_0" in m.instances
    assert "unit_0_1" in m.instances
    assert "unit_1_0" in m.instances
    assert "unit_1_1" in m.instances

    for inst in m.instances.values():
        assert inst.of is M
    assert len(m.signals) == 2
    assert "_unit_0_1_a_unit_0_0_b_" in m.signals
    assert "_unit_1_1_a_unit_1_0_b_" in m.signals

    ppkg = h.to_proto(m)
    assert isinstance(ppkg, h.proto.Package)


def test_instance_mult():
    """ Initial tests of instance-array-generation by multiplication """

    Child = h.Module(name="Child")
    Parent = h.Module(name="Parent")

    # Create an array via multiplication
    Parent.child = Child() * 5  # Lotta kids here

    # Check that the array is created correctly
    assert isinstance(Parent.child, h.InstArray)
    assert Parent.child.n == 5
    assert Parent.child.of is Child

    # Test elaboration
    h.elaborate(Parent)

    # Post-elaboration checks
    assert len(Parent.instances) == 5
    assert len(Parent.instarrays) == 0
    for inst in Parent.instances.values():
        assert inst.of is Child

    # Create another, this time via right-mul
    Parent = h.Module(name="Parent")
    Parent.child = 11 * Child()  # Getting even more kids

    # Check that the array is created correctly
    assert isinstance(Parent.child, h.InstArray)
    assert Parent.child.n == 11
    assert Parent.child.of is Child

    # Test elaboration
    h.elaborate(Parent)

    # Post-elaboration checks
    assert len(Parent.instances) == 11
    assert len(Parent.instarrays) == 0
    for inst in Parent.instances.values():
        assert inst.of is Child


def test_instance_mult2():
    """ Test connecting to multiplication-generated InstArrays """

    @h.module
    class Child:
        p = h.Port()

    @h.module
    class Parent:
        a = h.Signal(width=3)
        child = (Child() * 3)(p=a)

    assert len(Parent.instances) == 0
    assert len(Parent.instarrays) == 1

    h.elaborate(Parent)

    # Check that array-flattening completed correctly
    assert len(Parent.instances) == 3
    assert len(Parent.instarrays) == 0
    assert Parent.get("_child_0_").of is Child
    assert Parent.get("_child_1_").of is Child
    assert Parent.get("_child_2_").of is Child
    assert Parent.get("_child_0_").conns["p"] == Parent.a[0]
    assert Parent.get("_child_1_").conns["p"] == Parent.a[1]
    assert Parent.get("_child_2_").conns["p"] == Parent.a[2]


def test_instance_mult3():
    """ Test connecting to an "already connected" Instance """

    @h.module
    class Child:
        p = h.Port()

    @h.module
    class Parent:
        a = h.Signal(width=3)
        child = 3 * Child(p=a)  # <= this the one diff here from test_instance_mult2

    h.elaborate(Parent)

    # Check that array-flattening completed correctly
    assert len(Parent.instances) == 3
    assert len(Parent.instarrays) == 0
    assert Parent.get("_child_0_").of is Child
    assert Parent.get("_child_1_").of is Child
    assert Parent.get("_child_2_").of is Child
    assert Parent.get("_child_0_").conns["p"] == Parent.a[0]
    assert Parent.get("_child_1_").conns["p"] == Parent.a[1]
    assert Parent.get("_child_2_").conns["p"] == Parent.a[2]


def test_instance_mult4():
    """ Test connecting non-unit-width Arrays """

    @h.module
    class Child:
        a = h.Port(width=11)
        b = h.Port(width=16)

    @h.module
    class Parent:
        a = h.Signal(width=11)
        b = h.Signal(width=48)
        child = 3 * Child(a=a, b=b)

    h.elaborate(Parent)

    # Check that array-flattening completed correctly
    assert len(Parent.instances) == 3
    assert len(Parent.instarrays) == 0
    assert Parent.get("_child_0_").of is Child
    assert Parent.get("_child_1_").of is Child
    assert Parent.get("_child_2_").of is Child
    assert Parent.get("_child_0_").conns["a"] == Parent.a
    assert Parent.get("_child_1_").conns["a"] == Parent.a
    assert Parent.get("_child_2_").conns["a"] == Parent.a
    assert Parent.get("_child_0_").conns["b"] == Parent.b[0:15]
    assert Parent.get("_child_1_").conns["b"] == Parent.b[16:31]
    assert Parent.get("_child_2_").conns["b"] == Parent.b[32:47]


def test_netlist_fmts():
    """ Test netlisting basic types to several formats """

    @h.module
    class Bot:
        s = h.Input(width=3)
        p = h.Output()

    @h.module
    class Top:
        s = h.Signal(width=3)
        p = h.Output()
        b = Bot(s=s, p=p)

    # Convert to proto
    ppkg = h.to_proto(Top)
    assert len(ppkg.modules) == 2
    assert ppkg.modules[0].name == "hdl21.tests.test_hdl21.Bot"
    assert ppkg.modules[1].name == "hdl21.tests.test_hdl21.Top"

    # Convert the proto-package to a netlist
    nl = StringIO()
    h.netlist(ppkg, nl, "spectre")
    nl = nl.getvalue()

    # Basic checks on its contents
    assert "subckt Bot" in nl
    assert "+  s_2 s_1 s_0 p" in nl
    assert "subckt Top" in nl
    assert "b\n" in nl
    assert "+  ( s_2 s_1 s_0 p  )" in nl
    assert "+  p" in nl
    assert "+  Bot " in nl

    # Convert the proto-package to another netlist format
    nl = StringIO()
    h.netlist(ppkg, nl, "verilog")
    nl = nl.getvalue()

    # Basic checks on its contents
    assert "module Bot" in nl
    assert "input wire [2:0] s," in nl
    assert "output wire p" in nl
    assert "endmodule // Bot" in nl
    assert "module Top" in nl
    assert ".s(s)" in nl
    assert ".p(p)" in nl
    assert "endmodule // Top" in nl


def test_bad_width_conn():
    """ Test invalid connection-widths """
    c = h.Module(name="c")
    c.p = h.Port(width=3)  # Width-3 Port
    q = h.Module(name="q")
    q.s = h.Signal(width=5)  # Width-5 Signal
    q.c = c(p=q.s)  # <= Bad connection here

    with pytest.raises(RuntimeError):
        h.elaborate(q)


def test_bad_intf_conn():
    """ Test invalid Interface connections """

    @h.interface
    class P:
        p = h.Signal(width=3)

    @h.interface
    class R:
        z = h.Signal(width=11)

    c = h.Module(name="c")
    c.p = P(port=True)  # `P`-type Interface
    q = h.Module(name="q")
    q.r = R()  # `R`-type Interface
    q.c = c(p=q.r)  # <= Bad connection here

    with pytest.raises(RuntimeError):
        h.elaborate(q)


def test_array_concat_conn():
    """ Test connecting a `Concat` to an `InstArray` """

    Child = h.Module(name="Child")
    Child.p3 = h.Input(width=3)
    Child.p5 = h.Input(width=5)

    Parent = h.Module(name="Parent")
    Parent.s5 = h.Signal(width=5)
    Parent.s9 = h.Signal(width=9)
    Parent.c = 2 * Child()
    Parent.c.p3 = Parent.s5[0:2]
    Parent.c.p5 = h.Concat(Parent.s5[3], Parent.s9)

    import sys

    h.netlist(h.to_proto(Parent), sys.stdout, "verilog")


def test_slice_resolution():
    """ Test resolutions of slice combinations """

    from hdl21.elab import _resolve_slice

    # Slice of a Signal
    s = h.Signal(width=5)
    assert _resolve_slice(s[0]) == s[0]

    # Slice of a Concat
    s1 = h.Signal(name="s1", width=5)
    s2 = h.Signal(name="s2", width=5)
    c = h.Concat(s1, s2)
    assert _resolve_slice(c[0]) == s1[0]
    assert _resolve_slice(c[1]) == s1[1]
    assert _resolve_slice(c[2]) == s1[2]
    assert _resolve_slice(c[3]) == s1[3]
    assert _resolve_slice(c[4]) == s1[4]
    assert _resolve_slice(c[5]) == s2[0]
    assert _resolve_slice(c[6]) == s2[1]
    assert _resolve_slice(c[7]) == s2[2]
    assert _resolve_slice(c[8]) == s2[3]
    assert _resolve_slice(c[9]) == s2[4]

    # Slice of a Slice
    sa = h.Signal(name="sa", width=5)
    assert _resolve_slice(sa[2:5][0]) == sa[2]
    assert _resolve_slice(sa[2:5][1]) == sa[3]
    assert _resolve_slice(sa[2:5][2]) == sa[4]

    assert _resolve_slice(sa[:][0]) == sa[0]
    assert _resolve_slice(sa[:][1]) == sa[1]
    assert _resolve_slice(sa[:][2]) == sa[2]
    assert _resolve_slice(sa[:][3]) == sa[3]
    assert _resolve_slice(sa[:][4]) == sa[4]

    # Check a concatenation case. Note `Concat` does not support equality, so compare parts.
    assert _resolve_slice(sa[:][0:4]).parts == (sa[0], sa[1], sa[2], sa[3], sa[4])




"""
Hdl21 Parameters and Param-Classes 
"""

from typing import Optional, Any
import dataclasses
import json
import pickle
import hashlib
import pydantic


# Import a few likely common Exception-types,
# Not used directly here, but likely by users and tests
from pydantic import ValidationError
from dataclasses import FrozenInstanceError


class _Default:
    """ The cardinal value for unspecified parameter-declaration defaults.
    Normally the class-object itself would work for this, but `pydantic.dataclasses`
    seems to have some problems accepting class-objects. 
    https://github.com/samuelcolvin/pydantic/issues/1537
    So we create this singleton instance `_default`, AKA `_Default._the_one`. """

    _the_one = None

    def __new__(cls, *_, **__):
        if cls._the_one is None:
            cls._the_one = super().__new__(cls)
        return cls._the_one


# Create that singleton
_default = _Default()


def paramclass(cls: type) -> type:
    """ Parameter-Class Creation Decorator 

    Transforms a class-definition full of Params into a type-validated dataclass, 
    with methods for default value and description-dictionary retrieval. 

    Hdl21's `paramclass`es are immutable, strongly-typed data-storage structures. 
    They are defined through a syntax similar to `@dataclass`, but using the `Param` 
    constructor, and assignment rather than type annotation. 

    @paramclass
    class C:
        reqd = Param(dtype=int, desc="A Required Parameter")
        optn = Param(dtype=int, desc="An Optional Parameter", default=11)

    `Param`s each have required datatype (`dtype`) and description (`desc`) fields, 
    and optional default values. 
    Each `paramclass` constructor can be called with ordered arguments, 
    in the order defined in the `paramclass`, or with named arguments. 
    Named arguments are highly recommended for more than a single parameter. 
    Note Python's function-argument ordering requirements also dictate 
    that all `paramclass` required-arguments be declared *before* any optional arguments. 
    This also reinforces good practice for communicating which parameters are required.

    Each `paramclass` comes with class-methods `descriptions` and `defaults`, 
    which return dictionaries of the parameter names to descriptions and 
    names to default values (for those with defaults), respectively. 
    
    Requirements of the input `cls`:
    * *All* non-Python-internal fields must be of type `Param`
    * Inheritance is not supported
    """
    if cls.__bases__ != (object,):
        raise RuntimeError(f"Invalid @hdl21.paramclass inheriting from {cls.__bases__}")
    protected_names = ["descriptions", "defaults"]
    dunders = dict()
    params = dict()
    # Take a lap through the class dictionary, type-check everything and grab Params
    for key, val in cls.__dict__.items():
        if key in protected_names:
            raise RuntimeError(f"Invalid field name {key} in paramclass {cls}")
        elif key.startswith("__"):
            dunders[key] = val
        elif isinstance(val, Param):
            params[key] = val
        else:
            raise RuntimeError(
                f"Invalid class-attribute {key} in paramclass {cls}. All attributes should be `hdl21.Param`s."
            )
    # Translate the Params into dataclass.field-compatible tuples
    fields = list()
    for name, par in params.items():
        field = [name, par.dtype]
        if par.default is not _default:
            field.append(dataclasses.field(default=par.default))
        # Default factories: not supported, yet. See `Param` below.
        # elif par.default_factory is not _default:
        #     field.append(dataclasses.field(default_factory=par.default_factory))
        fields.append(tuple(field))
    # Add a few helpers to the class namespace
    ns = dict(
        __params__=params,
        __paramclass__=True,
        descriptions=classmethod(
            lambda cls: {k: v.desc for k, v in cls.__params__.items()}
        ),
        defaults=classmethod(
            lambda cls: {
                k: v.default
                for k, v in cls.__params__.items()
                if v.default is not _default
            }
        ),
    )
    # Create ourselves a (std-lib) dataclass
    cls = dataclasses.make_dataclass(cls.__name__, fields, namespace=ns, frozen=True)
    # Pass this through the pydantic dataclass-decorator-function
    cls = pydantic.dataclasses.dataclass(cls, frozen=True)
    # Pydantic seems to want to add this one *after* class-creation
    def _brick_subclassing_(cls, *_, **__):
        raise RuntimeError(
            f"Error: attempt to sub-class `hdl21.paramclass` {cls} is not supported"
        )

    cls.__init_subclass__ = classmethod(_brick_subclassing_)
    # And don't forget to return it!
    return cls


def isparamclass(cls: type) -> bool:
    """ Boolean indication of whether `cls` has been `@paramclass`-decorated """
    return getattr(cls, "__paramclass__", False)


@pydantic.dataclasses.dataclass
class Param:
    """ Parameter Declaration """

    dtype: Any  # Datatype. Required
    desc: str  # Description. Required
    default: Optional[Any] = _default  # Default Value. Optional

    # Default factories are supported in std-lib dataclasses, but "in beta" in `pydantic.dataclasses`.
    # default_factory: Optional[Any] = _default  # Default Call-Factory


def _unique_name(params: Any) -> str:
    """ Create a unique name for parameter-class instance `params` """
    if not isparamclass(params):
        raise RuntimeError(f"Invalid parameter-class instance {params}")

    # Determine whether *all* fields of `params` are scalar values: strings, numbers, and options thereof
    scalars = [
        str,
        int,
        float,
        type(None),
        Optional[str],
        Optional[int],
        Optional[float],
    ]
    # Boolean indication of whether *all* param-datatypes are from among these
    all_scalar = all([param.dtype in scalars for param in params.__params__.values()])

    # If all params are scalars, create a readable string of their values
    if all_scalar:
        name = params.__class__.__name__ + "("
        for pname in params.__params__.keys():
            pval = getattr(params, pname)
            name += pname + "=" + str(pval) + " "
        name = name.rstrip()
        name += ")"

        # These names must also be limited in length, for sake of our favorite output formats.
        # If the generated name is too long, use the hashing method below instead
        if len(name) < 128:  # Probably(?) a reasonable length limit
            return name

    # Non-scalar cases generally include nested `@paramclasses` or sequences,
    # or surpass the length-limits above. We serialize and hash them.
    # Preferably serialize as JSON
    jsonstr = json.dumps(params, indent=4, default=hdl21_naming_encoder)
    data = bytes(jsonstr, encoding="utf-8")

    # If JSON encoding fails, we *could* use pickle instead.
    # Disabled for now, because pickle is not guaranteed to be deterministic.
    # Note JSON is preferable for its run-to-run stability,
    # where pickle can generally serialize more types.
    # data = pickle.dumps(params)

    # Take that data and hash it
    # Note the "not used for security" option ensures consistent hashing between runs/ Python-processes
    h = hashlib.new("md5", usedforsecurity=False)
    h.update(data)
    # Combine the `@paramclass` name with this (hex) digest
    return params.__class__.__name__ + "(" + h.hexdigest() + ")"


def hdl21_naming_encoder(obj: Any) -> Any:
    """ JSON encoder for naming of Hdl21 parameter-values. 

    "Extends" `pydantic.json.pydantic_encoder` by first checking for 
    each of the non-serializable Hdl21 types (`Module`, `Instance`, `Generator`, etc.), 
    then hands everything else off to `pydantic.json.pydantic_encoder`. 

    Note this *does not fully serialize `Module`s and the like - 
    see `hdl21.to_proto` for this. This JSON-ization is just good enough 
    to enable unique naming of Hdl-object-value parameters. """

    from .module import Module, ExternalModule, ExternalModuleCall
    from .instance import Instance
    from .generator import Generator, GeneratorCall

    if isinstance(obj, (Instance, Generator)):
        # Not supported as parameters
        raise RuntimeError(f"Invalid `hdl21.paramclass` field {obj}")
    if isinstance(obj, (Module, ExternalModule)):
        # Modules use their qualified class names/paths
        return obj._qualname()
    if isinstance(obj, ExternalModuleCall):
        # Mix the qualified class names/paths with the parameters
        return obj.module._qualname() + _unique_name(obj.params)
    if isinstance(obj, GeneratorCall):
        # Most other Hdl21 objects as parameters are pending, maybe, maybe not, support as parameter-values.
        raise NotImplementedError

    # Dataclasses also require custom handling, as the default encoder deep-copies them,
    # often invoking methods not supported on several Hdl21 types.
    # Convert to (shallow) dictionaries instead.
    if dataclasses.is_dataclass(obj):
        return {f.name: getattr(obj, f.name) for f in dataclasses.fields(obj)}

    # Not an Hdl21 type. Hand off to pydantic.
    return pydantic.json.pydantic_encoder(obj)


# Shortcut for parameter-less generators.
# Defines the empty-value `@paramclass`.
HasNoParams = paramclass(
    dataclasses.make_dataclass("NoParams", fields=[], namespace={}, frozen=True)
)
# And an instance of it
NoParams = HasNoParams()



"""
hdl21 ProtoBuf Export
"""

from textwrap import dedent
from dataclasses import asdict, is_dataclass
from enum import Enum
from types import SimpleNamespace
from typing import Optional, List, Union

# Local imports
# Proto-definitions
from . import circuit_pb2 as protodefs

# HDL
from ..elab import Elabables, elab_all
from ..module import Module, ExternalModule, ExternalModuleCall
from ..primitives import Primitive, PrimitiveCall, PrimitiveType
from ..instance import Instance
from .. import signal


def to_proto(
    top: Elabables, domain: Optional[str] = None, **kwargs,
) -> protodefs.Package:
    """Convert Elaborate-able Module or Generator `top` and its dependencies to a Proto-format `Package`"""
    # Elaborate all the top-level Modules
    tops = elab_all(top)
    exporter = ProtoExporter(tops=tops, domain=domain)
    return exporter.export()


class ProtoExporter:
    """Hierarchical Protobuf Exporter.
    Walks a Module hierarchy from Module `self.top`, defining each required Module along the way.
    Modules are defined in `self.pkg` in dependency order.
    Upon round-tripping, all dependent child-modules will be encountered before their parent instantiators."""

    def __init__(self, tops: List[Module], domain: Optional[str] = None):
        self.tops = tops
        self.modules = dict()  # Module-id to Proto-Module dict
        self.module_names = dict()  # (Serialized) Module-name to Proto-Module dict
        self.ext_modules = dict()  # ExternalModule-id to Proto-ExternalModule dict
        # Default `domain` AKA package-name is the empty string
        self.pkg = protodefs.Package(domain=domain or "")

    def export(self) -> protodefs.Package:
        """Export starting with every Module in `self.tops`,
        visiting every hierarchical node along the way.
        Returns our generated `Package` as a result."""
        if not isinstance(self.tops, list):
            raise TypeError
        for m in self.tops:
            if not isinstance(m, Module):
                raise TypeError
            self.export_module(m)
        return self.pkg

    def export_module_name(self, module: Module) -> protodefs.QualifiedName:
        """Create and return a unique `QualifiedName` for Module `module`.
        Raises a `RuntimeError` if unique name is taken."""

        mname = module._qualname()
        if mname in self.module_names:
            conflict = self.module_names[mname]
            raise RuntimeError(
                dedent(
                    f"""\
                    Cannot serialize Module {module} due to conflicting name with {conflict}. 
                    (Was this a generator that didn't get decorated with `@hdl21.generator`?) """
                )
            )
        return mname

    def export_module(self, module: Module) -> protodefs.Module:
        if id(module) in self.modules:  # Already done
            return self.modules[id(module)]

        if module.interfaces:  # Can't handle these, at least for now
            raise RuntimeError(
                f"Invalid attribute for Proto export: Module {module.name} with Interfaces {list(module.interfaces.keys())}"
            )

        # Create the Proto-Module
        pmod = protodefs.Module()

        # Create its serialized name
        pmod.name = self.export_module_name(module)

        # Create its Port-objects
        for port in module.ports.values():
            pmod.ports.append(self.export_port(port))

        # Create its Signal-objects
        for sig in module.signals.values():
            psig = protodefs.Signal(name=sig.name, width=sig.width)
            pmod.signals.append(psig)

        # Create each Proto-Instance
        for inst in module.instances.values():
            if not inst._resolved:
                raise RuntimeError(
                    f"Invalid Instance {inst.name} of unresolved Module in Module {module.name}"
                )
            pinst = self.export_instance(inst)
            pmod.instances.append(pinst)

        # Store references to the result, and return it
        self.modules[id(module)] = pmod
        self.module_names[pmod.name] = pmod
        self.pkg.modules.append(pmod)
        return pmod

    def export_external_module(self, emod: ExternalModule) -> protodefs.ExternalModule:
        """Export an `ExternalModule`"""
        if id(emod) in self.ext_modules:  # Already done
            return self.ext_modules[id(emod)]

        # Create the Proto-ExternalModule
        qname = protodefs.QualifiedName(name=emod.name, domain=emod.domain)
        pmod = protodefs.ExternalModule(name=qname)

        # Create its Port-objects
        for port in emod.ports.values():
            pmod.ports.append(self.export_port(port))

        # Store references to the result, and return it
        self.ext_modules[id(emod)] = pmod
        self.pkg.ext_modules.append(pmod)
        return pmod

    @classmethod
    def export_primitive(cls, prim: Primitive) -> protodefs.ExternalModule:
        """Export a `Primitive as a `protodefs.ExternalModule`. 
        Not typically done as part of serialization, 
        but solely as an aid to other conversion utilities. """

        # Create the Proto-ExternalModule
        if prim.primtype == PrimitiveType.PHYSICAL:
            domain = "hdl21.primitives"
        elif prim.primtype == PrimitiveType.IDEAL:
            domain = "hdl21.ideal"
        else:
            raise ValueError
        qname = protodefs.QualifiedName(name=prim.name, domain=domain)
        pmod = protodefs.ExternalModule(name=qname)

        # Create its Port-objects
        for port in prim.port_list:
            pmod.ports.append(cls.export_port(port))

        # And return it
        return pmod

    @classmethod
    def export_port(cls, port: signal.Port) -> protodefs.Port:
        pport = protodefs.Port()
        pport.direction = cls.export_port_dir(port)
        pport.signal.name = port.name
        pport.signal.width = port.width
        return pport

    @classmethod
    def export_port_dir(cls, port: signal.Port) -> protodefs.Port.Direction:
        # Convert between Port-Direction Enumerations
        if port.direction == signal.PortDir.INPUT:
            return protodefs.Port.Direction.INPUT
        if port.direction == signal.PortDir.OUTPUT:
            return protodefs.Port.Direction.OUTPUT
        if port.direction == signal.PortDir.INOUT:
            return protodefs.Port.Direction.INOUT
        if port.direction == signal.PortDir.NONE:
            return protodefs.Port.Direction.NONE
        raise ValueError

    def export_instance(self, inst: Instance) -> protodefs.Instance:
        """Convert an hdl21.Instance into a Proto-Instance
        Depth-first retrieves a Module definition first,
        using its generated `name` field as the Instance's `module` pointer."""

        # Create the Proto-Instance
        pinst = protodefs.Instance(name=inst.name)

        # First depth-first seek out our definition,
        # Retrieving the data we need to make a `Reference` to it
        if isinstance(inst._resolved, Module):
            pmod = self.export_module(inst._resolved)
            # Give it a Reference to its Module
            pinst.module.local = pmod.name
        elif isinstance(inst._resolved, (PrimitiveCall, ExternalModuleCall)):
            call = inst._resolved
            if isinstance(inst._resolved, PrimitiveCall):
                # Create a reference to one of the `primitive` namespaces
                if call.prim.primtype == PrimitiveType.PHYSICAL:
                    pinst.module.external.domain = "hdl21.primitives"
                elif call.prim.primtype == PrimitiveType.IDEAL:
                    pinst.module.external.domain = "hdl21.ideal"
                else:
                    raise ValueError
                pinst.module.external.name = call.prim.name
                params = asdict(call.params)
            else:  # ExternalModuleCall
                self.export_external_module(call.module)
                pinst.module.external.domain = call.module.domain or ""
                pinst.module.external.name = call.module.name
                # FIXME: while these ExternalModule parameters can ostensibly be anything,
                # there are really two supported types-of-types:
                # dictionaries, and dataclasses (which we can turn into dictionaries)
                if isinstance(call.params, dict):
                    params = call.params
                elif is_dataclass(call.params):
                    params = asdict(call.params)
                else:
                    raise TypeError

            # Set the parameter-values
            for key, val in params.items():
                if isinstance(val, type(None)):
                    continue  # None-valued parameters go un-set
                elif isinstance(val, int):
                    pinst.parameters[key].integer = val
                elif isinstance(val, float):
                    pinst.parameters[key].double = val
                elif isinstance(val, str):
                    pinst.parameters[key].string = val
                elif isinstance(val, Enum):
                    # Enum-valued parameters are always strings
                    pinst.parameters[key].string = val.value
                else:
                    raise TypeError(f"Invalid instance parameter {val} for {inst}")
        else:
            raise TypeError

        # Create its connections mapping
        for pname, sig in inst.conns.items():
            # Assign each item into the connections dict.
            # The proto interface requires copying it along the way
            pinst.connections[pname].CopyFrom(self.export_conn(sig))

        return pinst

    def export_conn(
        self, sig: Union[signal.Signal, signal.Slice, signal.Concat]
    ) -> protodefs.Connection:
        """ Export a proto Connection """

        pconn = protodefs.Connection()  # Create a proto-Connection
        if isinstance(sig, signal.Signal):
            pconn.sig.name = sig.name
            pconn.sig.width = sig.width
        elif isinstance(sig, signal.Slice):
            pslice = self.export_slice(sig)
            pconn.slice.CopyFrom(pslice)
        elif isinstance(sig, signal.Concat):
            pconc = self.export_concat(sig)
            pconn.concat.CopyFrom(pconc)
        else:
            raise TypeError(f"Invalid argument to `ProtoExporter.export_conn`: {sig}")
        return pconn

    def export_slice(self, slize: signal.Slice) -> protodefs.Slice:
        """ Export a signal-`Slice`. 
        Fails if the parent is not a concrete `Signal`, 
        i.e. it is a `Concat` or another `Slice`. """
        if not isinstance(slize.signal, signal.Signal):
            msg = f"Export error: {slize} has a parent {slize.signal} which is not a concrete Signal"
            raise RuntimeError(msg)
        return protodefs.Slice(signal=slize.signal.name, top=slize.top, bot=slize.bot)

    def export_concat(self, concat: signal.Concat) -> protodefs.Concat:
        """ Export (potentially recursive) Signal Concatenations """
        pconc = protodefs.Concat()
        for part in concat.parts:
            pconc.parts.append(self.export_conn(part))
        return pconc



"""
# hdl21 Generators
"""

import inspect
import pickle
from dataclasses import field
from typing import Callable, Union, Any, Optional
from types import ModuleType
from pydantic.dataclasses import dataclass
from pydantic import ValidationError

# Local imports
from .module import Module
from .instance import calls_instantiate


@dataclass
class Generator:
    """ # Generator Object 
    Typically created by the `@hdl.generator` decorator. 
    Stores a function-object and parameters-type, 
    along with some auxiliary data. 
    """

    func: Callable  # The generator function
    paramtype: type  # Parameter type
    usecontext: bool  # Boolean indication of `Context` usage
    pymodule: ModuleType  # Python module where function defined

    def __call__(self, arg: Any):
        """ Calls to Generators create GeneratorCall-objects
        to be expanded during elaboration. """
        return GeneratorCall(gen=self, arg=arg)

    @property
    def Params(self) -> type:
        """ Parameter-Type Property """
        return self.paramtype


@calls_instantiate
@dataclass
class GeneratorCall:
    """ Generator 'Bare Calls' 
    Stored for expansion during elaboration. 
    Only single-argument calls with `Params` are supported. 
    Any application of a `Context` is done during elaboration.  """

    gen: Generator
    arg: Any
    result: Optional[Module] = field(init=False, default=None)

    def __post_init_post_parse__(self):
        """ Validate that our argument matches the Generator param-type """
        if not isinstance(self.arg, self.gen.Params):
            raise ValidationError

    def __eq__(self, other) -> bool:
        """ Generator-Call equality requires:
        * *Identity* between generators, and 
        * *Equality* between parameter-values. """
        return self.gen is other.gen and self.arg == other.arg

    def __hash__(self):
        """ Generator-Call hashing, consistent with `__eq__` above, uses:
        * *Identity* of its generator, and 
        * *Value* of its parameters. 
        The two are joined for hashing as a two-element tuple. """
        return hash((id(self.gen), pickle.dumps(self.arg)))


def generator(f: Callable) -> Generator:
    """ Decorator for Generator Functions """
    from .params import isparamclass
    from .elab import Context

    if not callable(f):
        raise RuntimeError(f"Invalid `@generator` application to non-callable {f}")

    # Grab, parse, and validate its call-signature
    sig = inspect.signature(f)

    args = list(sig.parameters.values())
    if len(args) < 1 or len(args) > 2:
        raise RuntimeError(f"Invalid generator call signature for {f.__name__}: {args}")

    # Extract the parameters-argument type
    paramtype = args[0].annotation
    if not isparamclass(paramtype):
        raise RuntimeError(
            f"Invalid generator call signature for {f.__name__}: {args}. First argument must be a paramclass."
        )
    # Check the context argument, if the function uses one
    usecontext = False
    if len(args) > 1:  # Also requests the context
        if args[1].annotation is not Context:
            raise RuntimeError(
                f"Invalid generator call signature for {f.__name__}: {args}. Second argument (if provided) must be a Context."
            )
        usecontext = True
    # Validate the return type is `Module`
    rt = sig.return_annotation
    if rt is not Module:
        raise TypeError(
            f"Generator {f.__name__} must return (and must be annotated to return) a Module."
        )

    # Grab a reference to the Python-module defining the function
    # (For later name-unique-ifying)
    pymodule = inspect.getmodule(f)
    return Generator(
        func=f, paramtype=paramtype, usecontext=usecontext, pymodule=pymodule
    )



"""
# hdl21 Connection Helpers 

Decorators which add a number of connection-related facilities onto classes to which they are applied. 
"""
from typing import Any, Union


def connectable(cls: type) -> type:
    """ Decorator for connectable types """
    cls.__connects__ = True
    return cls


def is_connectable(obj: Any) -> bool:
    """ Boolean indication of connect-ability """
    return getattr(obj, "__connects__", False)


def connects(cls: type) -> type:
    """ Decorator to add 'connect by call' and 'connect by setattr' semantics. 
    Applied to hdl21 internal types such as `Instance`, `InstArray` and `InterfaceInstance`. 
    
    `connects` classes have a few more subtle requirements, including that they 
    indicate when their constructors complete via an `_initialized` field, 
    include a `conns` connections-dict, and a `portrefs` dictionary of past port-references. """

    # First check and fail if any of the methods to be defined here are already defined elsewhere
    defined_here = ["__call__", "__setattr__", "__getattr__", "connect"]
    for key in defined_here:
        if key in cls.__dict__:
            raise RuntimeError(
                f"Invalid modification of {cls} with `@hdl21.connects`: {key} is already defined, and will not be over-written."
            )

    def __call__(self, **kwargs) -> "Self":
        """ Connect-by-call """
        for key, val in kwargs.items():
            self.connect(key, val)
        # Don't forget to retain ourselves at the call-site!
        return self

    def __setattr__(self, key: str, val: Any) -> None:
        """ Connect-by-setattr """
        if not getattr(self, "_initialized", False) or key.startswith("_"):
            # Bootstrapping phase: do regular setattrs to get started
            return object.__setattr__(self, key, val)
        if key in self.__getattribute__("_specialcases"):  # Special case(s)
            return object.__setattr__(self, key, val)
        self.connect(key, val)

    def connect(self, portname: str, signal: Union["Signal", "PortRef"]) -> "Self":
        """ Connect Signal `signal` to port (name) `portname`. 
        Called by both by-call and by-assignment convenience methods, and usable directly. 
        Direct calls to `connect` will generally be required for ports with otherwise illegal names, 
        e.g. Python language keywords (`in`, `from`, etc.), 
        or Hdl21 internal "keywords" (`name`, `ports`, `signals`, etc.). 
        Returns `self` to aid in method-chaining use-cases. """
        if not is_connectable(signal):
            raise TypeError(f"{self} attempting to connect non-connectable {signal}")
        self.conns[portname] = signal
        return self

    def __getattr__(self, key: str) -> Any:
        """ Port access by getattr """
        if not self.__getattribute__("_initialized") or key.startswith("_"):
            # Bootstrapping phase: do regular getattrs to get started
            return object.__getattribute__(self, key)

        if key in self.__getattribute__("_specialcases"):  # Special case(s)
            return object.__getattribute__(self, key)

        # After elaboration, the fancy PortRef creation below goes away. Only return ready-made connections.
        if self.__getattribute__("_elaborated"):
            conns = self.__getattribute__("conns")
            if key in conns.keys():
                return conns[key]
            raise AttributeError(f"No attribute {key} for {self}")

        # Return a `PortRef`, creating one if necessary
        return self.port_ref(key)

    def port_ref(self, key: str) -> "PortRef":
        """ Return a port-reference to name `key`, creating it if necessary. """
        from .instance import PortRef

        # Check in our existing port-references
        port_refs = self.__getattribute__("portrefs")
        if key in port_refs:
            return port_refs[key]

        # New reference; create, add, and return it
        port_ref = PortRef(inst=self, portname=key)
        port_refs[key] = port_ref
        return port_ref

    # Attach all of these to the class
    cls.__call__ = __call__
    cls.__setattr__ = __setattr__
    cls.__getattr__ = __getattr__
    cls.connect = connect
    cls.port_ref = port_ref

    # And don't forget to return it!
    return cls



"""
# hdl21.module 
The `Module` module (get it?)

This module primarily defines the core circuit-reuse element `hdl21.Module`. 
It also includes the fun machinery for class-syntax creation of `Module`s, 
particularly the `@module` (lower-case) decorator-function. 
"""

import inspect
from textwrap import dedent
from types import ModuleType
from typing import Any, Optional, List, Union, get_args, Tuple, Type
from pydantic.dataclasses import dataclass
from dataclasses import field

# Local imports
from .signal import Signal, Visibility
from .instance import calls_instantiate, Instance, InstArray
from .interface import InterfaceInstance

# Type-alias for HDL objects storable as `Module` attributes
ModuleAttr = Union[Signal, Instance, InstArray, InterfaceInstance]


def _is_module_attr(val: Any) -> bool:
    """ Boolean indication of whether `val` is a valid `hdl21.Module` attribute. """
    return isinstance(val, get_args(ModuleAttr))


@calls_instantiate
class Module:
    """
    # Module
    The central element of hardware re-use.
    """

    def __init__(self, *, name: Optional[str] = None):
        self.name = name
        self.ports = dict()
        self.signals = dict()
        self.instances = dict()
        self.instarrays = dict()
        self.interfaces = dict()
        self.namespace = dict()  # Combination of all these

        self._pymodule = _caller_pymodule()  #  (Python) module where called
        self._importpath = None  # Optional field set by importers
        self._moduledata = None  # Optional[ModuleData]
        self._updated = True  # Flag indicating whether we've been updated since creating `_moduledata`
        self._initialized = True

    def __setattr__(self, key: str, val: Any) -> None:
        """ Set-attribute over-ride, organizing into type-based containers """

        if not getattr(self, "_initialized", False) or key.startswith("_"):
            # Bootstrapping phase. Pass along to "regular" setattr.
            return super().__setattr__(key, val)

        # Protected attrs - the internal dict-names
        banned = [
            "ports",
            "signals",
            "instances",
            "instarrays",
            "interfaces",
            "namespace",
            "add",
            "get",
        ]
        if key in banned:
            raise RuntimeError(
                f"Error attempting to over-write protected attribute {key} of Module {self}"
            )
        # Special case(s)
        if key == "name":
            return super().__setattr__(key, val)

        if not _is_module_attr(val):
            raise TypeError(f"Invalid Module attribute {val} for {self}")

        # Checks out! Name `val` and add it to our type-based containers.
        val.name = key
        self._add(val)
        return None

    def add(self, val: ModuleAttr, *, name: Optional[str] = None) -> ModuleAttr:
        """ Add a named HDL object into one of our internal dictionaries.

        This allows for programmatic insertion of attributes whose names are not legal Python identifiers,
        such as keywords ('in', 'from') and those including invalid characters.

        The added object `val` is also provided as the return value, enabling chaining-style usages such as
        ```python
        instance.inp = MyModule.add(h.Input(name="in", width=5))
        ``` 
        and similar. """

        if not _is_module_attr(val):
            raise TypeError(f"Invalid Module attribute {val} for {self}")

        # Now sort out naming. We get two name-sources:
        # (a) the function-argument `name` and (b) the value's `name` attribute.
        # One or the other (and not both) must be set.
        if name is None and val.name is None:
            raise RuntimeError(
                f"Invalid anonymous attribute {val} cannot be added to Module {self.name}"
            )
        if name is not None and val.name is not None:
            raise RuntimeError(
                f"Conflicting names {name} and {val.name} cannot be added to Module {self.name}"
            )
        if name is not None:
            val.name = name

        # Now `val.name` is set appropriately.
        # Add it to our type-based containers, and return it.
        return self._add(val)

    def _add(self, val: ModuleAttr) -> ModuleAttr:
        """ Internal `add` and `setattr` implementation. 
        Primarily sort `val` into one of our type-based containers. """

        if isinstance(val, Signal):
            self.namespace[val.name] = val
            if val.vis == Visibility.PORT:
                self.ports[val.name] = val
            else:
                self.signals[val.name] = val
        elif isinstance(val, Instance):
            self.instances[val.name] = val
            self.namespace[val.name] = val
        elif isinstance(val, InstArray):
            self.instarrays[val.name] = val
            self.namespace[val.name] = val
        elif isinstance(val, InterfaceInstance):
            self.interfaces[val.name] = val
            self.namespace[val.name] = val
        else:
            raise TypeError(f"Invalid Module attribute {val} for {self}")

        # Give it a reference to us
        val._parent_module = self
        # And return our newly-added attribute
        return val

    def get(self, name: str) -> Optional[ModuleAttr]:
        """ Get module-attribute `name`. Returns `None` if not present. 
        Note unlike Python built-ins such as `getattr`, `get` returns solely 
        from the HDL namespace-worth of `ModuleAttr`s. """
        ns = self.__getattribute__("namespace")
        return ns.get(name, None)

    def __getattr__(self, key: str) -> Any:
        """ Include our namespace-worth of HDL objects in dot-access retrievals """
        if key.startswith("_"):
            return object.__getattribute__(self, key)
        ns = self.__getattribute__("namespace")
        if key in ns:
            return ns[key]
        return object.__getattribute__(self, key)

    def __init_subclass__(cls, *_, **__):
        """ Sub-Classing Disable-ization """
        raise RuntimeError(
            dedent(
                f"""\
                Error attempting to create {cls.__name__}
                Sub-Typing hdl21.Module is not supported. """
            )
        )

    def __repr__(self) -> str:
        if self.name:
            return f"Module(name={self.name})"
        return f"Module(_anon_)"

    @property
    def interface_ports(self) -> dict:
        """ Port-Exposed Interface Instances """
        return {name: intf for name, intf in self.interfaces.items() if intf.port}

    @property
    def io(self) -> dict:
        """ Combined Dictionary of `Signal`-valued and `Interface`-valued ports """
        rv = self.interface_ports
        rv.update(self.ports)
        return rv

    def _defpath(self) -> str:
        """ Helper for exporting. 
        Returns a string representing "where" this module was defined. 
        This is generally one of a few things: 
        * If "normally" defined via Python code, it's the Python module path 
        * If *imported*, it's the path inferred during import """
        if self._importpath:  # Imported. Return the period-separated import path.
            return ".".join(self._importpath)
        # Defined the old fashioned way. Use the Python module name.
        return self._pymodule.__name__

    def _qualname(self) -> Optional[str]:
        """ Helper for exporting. Returns the path-qualified name including 
        `_defpath` options above, and the `Module.name`. """
        if self.name is None:
            return None
        return self._defpath() + "." + self.name

    def __eq__(self, other: "Module") -> bool:
        if self.name is None or other.name is None:
            raise RuntimeError(f"Cannot invoke equality on unnamed Module {self}")
        return self._qualname() == other._qualname()

    def __hash__(self):
        if self.name is None:
            raise RuntimeError(f"Cannot invoke hashing on unnamed Module {self}")
        return hash(self._qualname())

    def __getstate__(self):
        if self.name is None:
            raise RuntimeError(f"Cannot invoke pickling on unnamed Module {self}")
        return self._qualname()

    @property
    def _data(self) -> "_ModuleData":
        """ Retrieve a `_ModuleData` corresponding to our namespace """
        if self._updated or self._moduledata is None:
            ns = self.__getattribute__("namespace")
            self._moduledata = _ModuleData(name=self.name, namespace=list(ns.values()))
        return self._moduledata


@dataclass(frozen=True)
class _ModuleData:
    """ Immutable data-class representation of `Module` content. 
    Designed for internal use by `Module`, particularly for sake of 
    hashing, equality testing, and serialization. """

    qualname: str
    namespace: Tuple  # More specifically Tuple[ModuleAttr], but specifying so triggers a pydantic bug


def module(cls: type) -> Module:
    """
    # Module Definition Decorator

    Converts a class-body full of Module-storable attributes to an `hdl21.Module`.
    Example Usage:

    ```python
    import hdl21 as h

    @h.module
    class M1:
        d = h.Port()
        e = h.Signal()

    @h.module
    class M2:
        q = h.Signal()
        i = M1(d=q)
    ```

    `hdl21.Modules` are strongly-typed containers of other hardware objects:
    * `Signals` and `Ports`
    * Instances of other `Modules`
    * Structured connections via `Interfaces`

    Attempts to set module attributes of any other types will raise a `TypeError`.
    Notably this includes any behavioral elements implemented by Python functions.
    If you'd like arbitrary-executing Python-code that creates or manipulates `Modules`,
    you probably want an `hdl21.Generator` instead.

    Temporary variables can be declared inside of the `@module`-decorated class-body,
    so long as they are meet the Python convention for private data: name-prefixing with an underscore.
    These temporary variables are *not* propagated along as members of the `Module`.
    """

    if cls.__bases__ != (object,):
        raise RuntimeError(f"Invalid @hdl21.module inheriting from {cls.__bases__}")

    # Create the Module object
    module = Module(name=cls.__name__)

    dunders = dict()
    unders = dict()

    # Take a lap through the class dictionary, type-check everything and assign relevant attributes to the interface
    for key, val in cls.__dict__.items():
        if key.startswith("__"):
            dunders[key] = val
        elif key.startswith("_"):
            unders[key] = val
        else:
            setattr(module, key, val)
    # And return the Module
    return module


@dataclass
class ExternalModule:
    """
    # External Module

    Wrapper for circuits defined outside Hdl21, such as:
    * Inclusion of existing SPICE or Verilog netlists
    * Foundry or technology-specific primitives

    Unlike `Modules`, `ExternalModules` include parameters to support legacy HDLs.
    Said parameters may only take on a limited number of datatypes, and may not be nested.
    Each `ExternalModule` stores a parameter-type field `paramtype`. 
    Parameter-values are checked to be instances of `paramtype` at creation time. 
    """

    name: str
    port_list: List[Signal]
    paramtype: Type = object
    desc: Optional[str] = None  # Description
    domain: Optional[str] = None
    pymodule: Optional[ModuleType] = field(repr=False, init=False, default=None)
    importpath: Optional[List[str]] = field(repr=False, init=False, default=None)

    def __post_init__(self):
        # Internal tracking data: defining module/import-path
        self.pymodule = _caller_pymodule()
        self.importpath = None

    def __post_init_post_parse__(self):
        """After type-checking, do some more checks on values"""
        for p in self.port_list:
            if not p.name:
                raise ValueError(f"Unnamed Primitive Port {p} for {self.name}")
            if p.vis != Visibility.PORT:
                raise ValueError(
                    f"Invalid Primitive Port {p.name} on {self.name}; must have PORT visibility"
                )

    def __call__(self, params) -> "ExternalModuleCall":
        return ExternalModuleCall(module=self, params=params)

    @property
    def ports(self) -> dict:
        return {p.name: p for p in self.port_list}

    def _defpath(self) -> str:
        """ Helper for exporting. 
        Returns a string representing "where" this module was defined. 
        This is generally one of a few things: 
        * If "normally" defined via Python code, it's the Python module path 
        * If *imported*, it's the path inferred during import """
        if self.importpath:  # Imported. Return the period-separated import path.
            return ".".join(self.importpath)
        # Defined the old fashioned way. Use the Python module name.
        return self.pymodule.__name__

    def _qualname(self) -> Optional[str]:
        """ Helper for exporting. Returns the path-qualified name including 
        `_defpath` options above, and the `Module.name`. """
        if self.name is None:
            return None
        return self._defpath() + "." + self.name

    def __eq__(self, other: "ExternalModule") -> bool:
        if self.name is None or other.name is None:
            raise RuntimeError(f"Cannot invoke equality on unnamed Module {self}")
        return self._qualname() == other._qualname()

    def __hash__(self):
        if self.name is None:
            raise RuntimeError(f"Cannot invoke hashing on unnamed Module {self}")
        return hash(self._qualname())

    def __getstate__(self):
        if self.name is None:
            raise RuntimeError(f"Cannot invoke pickling on unnamed Module {self}")
        return self._qualname()


@calls_instantiate
@dataclass
class ExternalModuleCall:
    """ External Module Call
    A combination of an `ExternalModule` and its Parameter-values,
    typically generated by calling the Module. """

    module: ExternalModule
    params: Any

    def __post_init_post_parse__(self):
        # Type-validate our parameters
        if not isinstance(self.params, self.module.paramtype):
            raise TypeError(
                f"Invalid parameters {self.params} for ExternalModule {self.module}. Must be {self.module.paramtype}"
            )

    @property
    def ports(self) -> dict:
        return self.module.ports


def _caller_pymodule():
    """ Find the first frame not from *this* file
    Sometimes this will be a Generator. That's OK, they'll figure it out. """
    for fr in inspect.stack():
        # Note frames produce an `Optional[ModuleType]`.
        # (It seems extension modules may not have one.)
        # So be sure to check for `None`!
        pymod = inspect.getmodule(fr[0])
        if pymod is not None and pymod.__file__ != __file__:
            return pymod
    raise RuntimeError("Could not find caller module")



""" 
# Hdl21 Netlisting 

Exports protodefs.Package to a netlist format.
"""

# Std-Lib Imports
from typing import Optional, Tuple, List, Mapping, Union, IO
from enum import Enum
from dataclasses import field

# PyPi
from pydantic.dataclasses import dataclass

# Local Imports
from ..proto.to_proto import ProtoExporter
from ..proto.from_proto import ProtoImporter
from ..proto import circuit_pb2 as protodefs

# Internal type shorthand
ModuleLike = Union[protodefs.Module, protodefs.ExternalModule]


class NetlistFormat(Enum):
    """ Enumeration of available formats. 
    Includes string-value conversion/ generation. """

    SPECTRE = "spectre"
    VERILOG = "verilog"

    @staticmethod
    def get(spec: "NetlistFormatSpec") -> "NetlistFormat":
        """ Get the format specified by `spec`, in either enum or string terms. 
        Only does real work in the case when `spec` is a string, otherwise returns it unchanged. """
        if isinstance(spec, (NetlistFormat, str)):
            return NetlistFormat(spec)
        raise TypeError

    def netlister(self) -> type:
        """ Get the paired netlister-class """
        if self == NetlistFormat.SPECTRE:
            return SpectreNetlister
        if self == NetlistFormat.VERILOG:
            return VerilogNetlister
        raise ValueError


# Type-alias for specifying format, either in enum or string terms
NetlistFormatSpec = Union[NetlistFormat, str]


@dataclass
class Indent:
    """ Indentation Helper """

    chars: str = "  "
    num: int = field(init=False, default=0)
    state: str = field(init=False, default="")

    def __post_init_post_parse__(self) -> None:
        self.state = self.chars * self.num

    def __iadd__(self, other: int) -> None:
        """ In-place add, i.e. `indent += 1` """
        self.num += other
        self.state = self.chars * self.num
        return self

    def __isub__(self, other: int) -> None:
        """ In-place subtract, i.e. `indent -= 1` """
        self.num = self.num - other
        if self.num < 0:
            raise ValueError("Negative indentation")
        self.state = self.chars * self.num
        return self


class Netlister:
    """ # Abstract Base `Netlister` Class 

    `Netlister` is not directly instantiable, and none of its sub-classes are intended 
    for usage outside `hdl21.netlist`. The primary API method `netlist` is designed to 
    create, use, and drop a `Netlister` instance. 
    Once instantiated a `Netlister`'s primary API method is `netlist`. 
    This writes all content in its `pkg` field to destination `dest`. 
    
    Internal methods come in two primary flavors:
    * `write_*` methods, which write to `self.dest`. These methods are generally format-specific. 
    * `format_*` methods, which return format-specific strings, but *do not* write to `dest`. 
    * `get_*` methods, which retrieve some internal data, e.g. extracting the type of a `Connection`. 
    """

    enum = None  # Class attr of the `NetlistFormat` enum

    def __init__(self, pkg: protodefs.Package, dest: IO):
        self.pkg = pkg
        self.dest = dest
        self.indent = Indent(chars="  ")
        self.module_names = set()  # Netlisted Module names
        self.pmodules = dict()  # Visited proto-Modules
        self.ext_modules = dict()  # Visited ExternalModules
        self.ext_module_names = (
            dict()
        )  # Visited ExternalModule names, checked for duplicates

    def netlist(self) -> None:
        """ Primary API Method.
        Convert everything in `self.pkg` and write to `self.dest`. """

        if self.pkg.ext_sources:
            raise RuntimeError(f"External sources not (yet) supported")

        # First visit any externally-defined Modules,
        # Ensuring we have their port-orders.
        for emod in self.pkg.ext_modules:
            self.get_external_module(emod)

        # Now do the real stuff,
        # Creating netlist entries for each package-defined Module
        for mod in self.pkg.modules:
            self.write_module_definition(mod)
        # And ensure all output makes it to `self.dest`
        self.dest.flush()

    def write(self, s: str) -> None:
        """ Helper/wrapper, passing to `self.dest` """
        self.dest.write(s)

    def writeln(self, s: str) -> None:
        """ Write `s` as a line, at our current `indent` level. """
        self.write(f"{self.indent.state}{s}\n")

    def get_external_module(self, emod: protodefs.ExternalModule) -> None:
        """ Visit an ExternalModule definition.
        "Netlisting" these doesn't actually write anything,
        but just stores a reference  in internal dictionary `ext_modules`
        for future references to them. """
        key = (emod.name.domain, emod.name.name)
        if key in self.ext_modules:
            raise RuntimeError(f"Invalid doubly-defined external module {emod}")
        self.ext_modules[key] = emod

    @classmethod
    def get_param_default(cls, pparam: protodefs.Parameter) -> Optional[str]:
        """ Get the default value of `pparam`. Returns `None` for no default. """
        if pparam.WhichOneof("value") is None:
            return None
        return cls.get_param_value(pparam)

    @classmethod
    def get_param_value(cls, pparam: protodefs.Parameter) -> str:
        """ Get a string representation of a parameter-value """
        ptype = pparam.WhichOneof("value")
        if ptype == "integer":
            return str(int(pparam.integer))
        if ptype == "double":
            return str(float(pparam.double))
        if ptype == "string":
            return str(pparam.string)
        raise ValueError

    def get_module_name(self, module: protodefs.Module) -> str:
        """ Create a netlist-compatible name for proto-Module `module` """

        # Create the module name
        # Replace all format-invalid characters with underscores
        name = module.name.split(".")[-1]
        for ch in name:
            if not (ch.isalpha() or ch.isdigit() or ch == "_"):
                name = name.replace(ch, "_")
        return name

    def resolve_reference(self, ref: protodefs.Reference) -> Tuple[ModuleLike, str]:
        """ Resolve the `ModuleLike` referent of `ref`, along with its name. """

        if ref.WhichOneof("to") == "local":  # Internally-defined Module
            module = self.pmodules.get(ref.local, None)
            if module is None:
                raise RuntimeError(f"Invalid undefined Module {ref.local} ")
            module_name = self.get_module_name(module)

        elif ref.WhichOneof("to") == "external":  # Defined outside package

            # First check the priviledged/ internally-defined domains
            if ref.external.domain == "hdl21.primitives":
                msg = f"Invalid direct-netlisting of physical `hdl21.Primitive` `{ref.external.name}`. "
                msg += "Either compile to a target technology, or replace with an `ExternalModule`. "
                raise RuntimeError(msg)
            if ref.external.domain == "hdl21.ideal":
                # Ideal elements
                name = ref.external.name

                # Sort out the spectre-format name
                if name == "IdealCapacitor":
                    module_name = "capacitor"
                elif name == "IdealResistor":
                    module_name = "resistor"
                elif name == "IdealInductor":
                    module_name = "inductor"
                elif name == "VoltageSource":
                    module_name = "vsource"
                else:
                    raise ValueError(f"Unsupported or Invalid Ideal Primitive {ref}")

                # Awkardly, primitives don't naturally have definitions as
                # either `protodefs.Module` or `protodefs.ExternalModule`.
                # So we create one on the fly.
                prim = ProtoImporter.import_primitive(ref.external)
                module = ProtoExporter.export_primitive(prim)

            else:  # External Module
                key = (ref.external.domain, ref.external.name)
                module = self.ext_modules.get(key, None)
                if module is None:
                    msg = f"Invalid Instance of undefined External Module {key}"
                    raise RuntimeError(msg)
                # Check for duplicate names which would conflict from other namespaces
                module_name = ref.external.name
                if (
                    module_name in self.ext_module_names
                    and self.ext_module_names[module_name] is not module
                ):
                    msg = f"Conflicting ExternalModule definitions {module} and {self.ext_module_names[module_name]}"
                    raise RuntimeError(msg)
                self.ext_module_names[module_name] = module

        else:  # Not a Module, not an ExternalModule, not sure what it is
            raise ValueError

        return module, module_name

    def format_connection(self, pconn: protodefs.Connection) -> str:
        """ Format a `Connection` reference. 
        Does not *declare* any new connection objects, but generates references to existing ones. """
        # Connections are a proto `oneof` union
        # which includes signals, slices, and concatenations.
        # Figure out which to import

        stype = pconn.WhichOneof("stype")
        if stype == "sig":
            return self.format_signal_ref(pconn.sig)
        if stype == "slice":
            return self.format_signal_slice(pconn.slice)
        if stype == "concat":
            return self.format_concat(pconn.concat)
        raise ValueError(f"Invalid Connection Type {stype} for Connection {pconn}")

    """ 
    Virtual `format` Methods 
    """

    @classmethod
    def format_param_decl(cls, name: str, param: protodefs.Parameter) -> str:
        """ Format a named `Parameter` definition """
        raise NotImplementedError

    @classmethod
    def format_port_decl(cls, pport: protodefs.Port) -> str:
        """ Format a declaration of a `Port` """
        raise NotImplementedError

    @classmethod
    def format_port_ref(cls, pport: protodefs.Port) -> str:
        """ Format a reference to a `Port` """
        raise NotImplementedError

    @classmethod
    def format_signal_decl(cls, psig: protodefs.Signal) -> str:
        """ Format a declaration of Signal `psig` """
        raise NotImplementedError

    @classmethod
    def format_signal_ref(cls, psig: protodefs.Signal) -> str:
        """ Format a reference to Signal `psig` """
        raise NotImplementedError

    @classmethod
    def format_signal_slice(cls, pslice: protodefs.Slice) -> str:
        """ Format Signal-Slice `pslice` """
        raise NotImplementedError

    def format_concat(self, pconc: protodefs.Concat) -> str:
        """ Format the Concatenation of several other Connections """
        raise NotImplementedError

    @classmethod
    def format_bus_bit(cls, index: Union[int, str]) -> str:
        """ Format bus-bit `index` """
        raise NotImplementedError

    """ 
    Virtual `write` Methods 
    """

    def write_module_definition(self, pmodule: protodefs.Module) -> None:
        """ Write Module `module` """
        raise NotImplementedError

    def write_instance(self, pinst: protodefs.Instance) -> str:
        """ Write Instance `pinst` """
        raise NotImplementedError


class SpectreNetlister(Netlister):
    """Netlister for Spectre compatible netlist"""

    enum = NetlistFormat.SPECTRE  # Class attr of the `NetlistFormat` enum

    def write_module_definition(self, module: protodefs.Module) -> None:
        """ Create a Spectre-format definition for proto-Module `module` """

        # Create the module name
        module_name = self.get_module_name(module)
        # Check for double-definition
        if module_name in self.module_names:
            raise RuntimeError(f"Module {module_name} doubly defined")
        # Add to our visited lists
        self.module_names.add(module_name)
        self.pmodules[module.name] = module

        # Create the sub-circuit definition header
        self.write(f"subckt {module_name} \n")

        if module.ports:  # Create its ports
            self.write("+  ")
            for pport in module.ports:
                self.write(self.format_port_decl(pport) + " ")
            self.write("\n")
        else:
            self.write("+  // No ports \n")

        # Create its parameters, if defined
        if module.default_parameters:
            self.write("parameters ")
            for name, pparam in module.parameters.items():
                self.write(self.format_param_decl(name, pparam))
            self.write("\n")
        else:
            self.write("+  // No parameters \n")

        self.write("\n")  # End "header" facets
        # Note nothing need be done for internal signals;
        # spice and spectre create these "out of thin air"

        # Create its instances
        for pinst in module.instances:
            self.write_instance(pinst)
        self.write("\n")

        # Close up the sub-circuit
        self.write("ends \n\n")

    def write_instance(self, pinst: protodefs.Instance) -> None:
        """Create and return a netlist-string for Instance `pinst`"""

        # Create the instance name
        self.write(pinst.name + "\n")

        # Get its Module or ExternalModule definition, primarily for sake of port-order
        module, module_name = self.resolve_reference(pinst.module)

        if module.ports:
            self.write("+  ( ")
            # Get `module`'s port-order
            port_order = [pport.signal.name for pport in module.ports]
            # And write the Instance ports, in that order
            for pname in port_order:
                pconn = pinst.connections.get(pname, None)
                if pconn is None:
                    raise RuntimeError(f"Unconnected Port {pname} on {pinst.name}")
                self.write(self.format_connection(pconn) + " ")
            self.write(" ) \n")
        else:
            self.write("+  // No ports \n")

        # Write the module-name
        self.write("+  " + module_name + " \n")

        if pinst.parameters:  # Write the parameter-values
            self.write("+  ")
            for pname, pparam in pinst.parameters.items():
                pval = self.get_param_value(pparam)
                self.write(f"{pname}={pval} ")
            self.write(" \n")
        else:
            self.write("+  // No parameters \n")

        # And add a post-instance blank line
        self.write("\n")

    def format_concat(self, pconc: protodefs.Concat) -> str:
        """ Format the Concatenation of several other Connections """
        out = ""
        for part in pconc.parts:
            out += self.format_connection(part) + " "
        return out

    @classmethod
    def format_port_decl(cls, pport: protodefs.Port) -> str:
        """ Get a netlist `Port` definition """
        return cls.format_signal_ref(pport.signal)

    @classmethod
    def format_port_ref(cls, pport: protodefs.Port) -> str:
        """ Get a netlist `Port` reference """
        return cls.format_signal_ref(pport.signal)

    @classmethod
    def format_signal_ref(cls, psig: protodefs.Signal) -> str:
        """ Get a netlist definition for Signal `psig` """
        if psig.width < 1:
            raise RuntimeError
        if psig.width == 1:  # width==1, i.e. a scalar signal
            return psig.name
        # Vector/ multi "bit" Signal. Creates several spice signals.
        return " ".join(
            [f"{psig.name}{cls.format_bus_bit(k)}" for k in reversed(range(psig.width))]
        )

    @classmethod
    def format_signal_slice(cls, pslice: protodefs.Slice) -> str:
        """Get a netlist definition for Signal-Slice `pslice`"""
        base = pslice.signal
        indices = list(reversed(range(pslice.bot, pslice.top + 1)))
        if not len(indices):
            raise RuntimeError(f"Attempting to netlist empty slice {pslice}")
        return " ".join([f"{base}{cls.format_bus_bit(k)}" for k in indices])

    @classmethod
    def format_bus_bit(cls, index: Union[int, str]) -> str:
        """ Format-specific string-representation of a bus bit-index"""
        # Spectre netlisting uses an underscore prefix, e.g. `bus_0`
        return "_" + str(index)


class VerilogNetlister(Netlister):
    """ Netlister to create structural verilog netlists """

    enum = NetlistFormat.VERILOG  # Class attr of the `NetlistFormat` enum

    def write_module_definition(self, module: protodefs.Module) -> None:
        """ Create a Verilog module definition for proto-Module `module` """

        # Create the module name
        module_name = self.get_module_name(module)
        # Check for double-definition
        if module_name in self.module_names:
            raise RuntimeError(f"Module {module_name} doubly defined")
        # Add to our visited lists
        self.module_names.add(module_name)
        self.pmodules[module.name] = module

        # Create the module header
        self.writeln(f"module {module_name}")

        # Create its parameters, if defined
        if module.default_parameters:
            self.writeln("#( ")
            self.indent += 1
            for num, name in enumerate(module.default_parameters):
                pparam = module.default_parameters[name]
                comma = "" if num == len(module.default_parameters) - 1 else ","
                self.writeln(self.format_param_decl(name, pparam) + comma)
            self.indent -= 1
            self.writeln(") ")
        else:
            self.writeln("// No parameters ")

        if module.ports:  # Create its ports
            # Don't forget, a trailing comma after the last one is fatal to high-tech Verilog parsers!
            self.writeln("( ")
            self.indent += 1
            for num, pport in enumerate(module.ports):
                comma = "" if num == len(module.ports) - 1 else ","
                self.writeln(self.format_port_decl(pport) + comma)
            self.indent -= 1
            self.writeln("); ")
        else:
            self.writeln("// No ports ")

        self.writeln("")  # Blank line to end "header" facets
        self.indent += 1

        if module.signals:  # Create Signal declarations
            self.writeln("// Signal Declarations")
            for psig in module.signals:
                self.writeln(self.format_signal_decl(psig) + "; ")
        else:
            self.writeln("// No Signal Declarations")

        if module.instances:  # Create its instances
            self.writeln("")
            self.writeln("// Instance Declarations")
            for pinst in module.instances:
                self.write_instance(pinst)
        else:
            self.writeln("// No Instances")

        # Close up the module
        self.indent -= 1
        self.writeln("")  # Blank before `endmodule`
        self.writeln(f"endmodule // {module_name} \n\n")

    def write_instance(self, pinst: protodefs.Instance) -> None:
        """ Format and write Instance `pinst` """

        # Get its Module or ExternalModule definition
        module, module_name = self.resolve_reference(pinst.module)

        # Write the module-name
        self.writeln(module_name)

        if pinst.parameters:  # Write the parameter-values
            self.writeln("#( ")
            self.indent += 1
            for pname, pparam in pinst.parameters.items():
                pval = self.get_param_value(pparam)
                self.writeln(f"{pname}={pval} ")
            self.indent -= 1
            self.writeln(") ")
        else:
            self.writeln("// No parameters ")

        # Write the instance name
        self.writeln(pinst.name)

        if module.ports:  # Write connections, by-name, in-order
            self.writeln("( ")
            self.indent += 1
            # Get `module`'s port-order
            port_order = [pport.signal.name for pport in module.ports]
            # And write the Instance ports, in that order
            for num, pname in enumerate(port_order):
                pconn = pinst.connections.get(pname, None)
                if pconn is None:
                    raise RuntimeError(f"Unconnected Port {pname} on {pinst.name}")
                # Again a trailing comma after the last one is fatal!
                comma = "" if num == len(port_order) - 1 else ","
                self.writeln(f".{pname}({self.format_connection(pconn)}){comma} ")
            # Close up the ports
            self.indent -= 1
            self.writeln("); ")
        else:
            self.writeln("// No ports ")

        self.writeln("")  # Post-Instance blank line

    @classmethod
    def format_param_type(cls, pparam: protodefs.Parameter) -> str:
        """ Verilog type-string for `Parameter` `param`. """
        ptype = pparam.WhichOneof("value")
        if ptype == "integer":
            return "longint"
        if ptype == "double":
            return "real"
        if ptype == "string":
            return "string"
        raise ValueError

    @classmethod
    def format_param_decl(cls, name: str, param: protodefs.Parameter) -> str:
        """ Format a parameter-declaration """
        rv = f"parameter {name}"
        # FIXME: whether to include datatype
        # dtype = cls.format_param_type(param)
        default = cls.get_param_default(param)
        if default is not None:
            rv += f" = {default}"
        return rv

    def format_concat(self, pconc: protodefs.Concat) -> str:
        """ Format the Concatenation of several other Connections """
        # Verilog { a, b, c } concatenation format
        parts = [self.format_connection(part) for part in pconc.parts]
        return "{" + ", ".join(parts) + "}"

    @classmethod
    def format_port_decl(cls, pport: protodefs.Port) -> str:
        """ Format a `Port` declaration """

        # First retrieve and check the validity of its direction
        port_type_to_str = {
            protodefs.Port.Direction.Value("INPUT"): "input",
            protodefs.Port.Direction.Value("OUTPUT"): "output",
            protodefs.Port.Direction.Value("INOUT"): "inout",
            protodefs.Port.Direction.Value("NONE"): "NO_DIRECTION",
        }
        dir_ = port_type_to_str.get(pport.direction, None)
        if dir_ is None:
            msg = f"Invalid Verilog netlisting for unknown Port direction {pport.direction}"
            raise RuntimeError(msg)
        if dir_ == "NO_DIRECTION":
            msg = f"Invalid Verilog netlisting for undirected Port {pport}"
            raise RuntimeError(msg)

        return dir_ + " " + cls.format_signal_decl(pport.signal)

    @classmethod
    def format_signal_decl(cls, psig: protodefs.Signal) -> str:
        """ Format a `Signal` declaration """
        rv = "wire"
        if psig.width > 1:
            rv += f" [{psig.width-1}:0]"
        rv += f" {psig.name}"
        return rv

    @classmethod
    def format_port_ref(cls, pport: protodefs.Port) -> str:
        """ Format a reference to a `Port`. 
        Unlike declarations, this just requires the name of its `Signal`. """
        return cls.format_signal_ref(pport.signal)

    @classmethod
    def format_signal_ref(cls, psig: protodefs.Signal) -> str:
        """ Format a reference to a `Signal`. 
        Unlike declarations, this just requires its name. """
        return psig.name

    @classmethod
    def format_signal_slice(cls, pslice: protodefs.Slice) -> str:
        """ Format Signal-Slice `pslice` """
        if pslice.top == pslice.bot:  # Single-bit slice
            return f"{pslice.signal}[{pslice.top}]"
        return f"{pslice.signal}[{pslice.top}:{pslice.bot}]"  # Multi-bit slice


def netlist(
    pkg: protodefs.Package, dest: IO, fmt: NetlistFormatSpec = "spectre"
) -> None:
    """ Netlist proto-Package `pkg` to destination `dest`. 

    Example usages: 
    ```python
    h.netlist(pkg, dest=open('mynetlist.v', 'w'), fmt='verilog')
    ```
    ```python
    s = StringIO()
    h.netlist(pkg, dest=s, fmt='spectre')
    ```
    ```python
    import sys
    h.netlist(pkg, dest=sys.stdout, fmt='spice')
    ```

    Primary argument `pkg` must be a `protodefs.Package`.  
    Destination `dest` may be anything that supports the `typing.IO` interface, 
    commonly including open file-handles. `StringIO` is particularly helpful 
    for producing a netlist in an in-memory string.  
    Format-specifier `fmt` may be any of the `NetlistFormatSpec` enumerated values 
    or their string equivalents.
    """
    fmt_enum = NetlistFormat.get(fmt)
    netlister_cls = fmt_enum.netlister()
    netlister = netlister_cls(pkg, dest)
    netlister.netlist()




"""
hdl21 ProtoBuf Import 
"""
from types import SimpleNamespace
from typing import Union, Any, Dict, List, Type

# Local imports
# Proto-definitions
from . import circuit_pb2 as protodefs

# HDL
from ..module import Module, ExternalModule
from ..instance import Instance
from ..signal import Signal, Port, PortDir, Slice, Concat
from .. import primitives
from ..primitives import Primitive


def from_proto(pkg: protodefs.Package) -> SimpleNamespace:
    """ Convert Proto-defined Package `pkg` to a namespace-full of Modules. """
    importer = ProtoImporter(pkg)
    return importer.import_()


class ProtoImporter:
    """ Protobuf Package Importer. 
    Collects all `Modules` defined in Protobuf-sourced primary-argument `pkg` into a Python `types.SimpleNamespace`. """

    def __init__(self, pkg: protodefs.Package):
        self.pkg = pkg
        self.modules = dict()  # Dict of names to Modules
        self.ext_modules = dict()  # Dict of qual-names to ExternalModules
        self.ns = SimpleNamespace()
        self.ns.name = pkg.domain

    def import_(self) -> SimpleNamespace:
        """ Import the top-level `Package` to a Python namespace """
        # Walk through each proto-defined Module
        # External modules first, as we know they have no dependencies
        for emod in self.pkg.ext_modules:
            self.import_external_module(emod)
        for pmod in self.pkg.modules:
            self.import_module(pmod)
        # Return our collected namespace
        return self.ns

    def get_namespace(self, path: List[str]) -> SimpleNamespace:
        # Get a (potentially nested) namespace at `path`,
        # creating levels along the way if necessary.
        ns = self.ns
        for part in path:
            attr = getattr(ns, part, None)
            if attr is None:  # Create a new Namespce
                new_ns = SimpleNamespace()
                new_ns.name = part
                setattr(ns, part, new_ns)
                ns = new_ns
            elif isinstance(attr, SimpleNamespace):
                ns = attr
            else:
                raise RuntimeError(f"Invalid namespace path {path} overwriting {attr}")
        return ns

    def import_external_module(self, pmod: protodefs.ExternalModule) -> ExternalModule:
        """ Convert Proto-Module `emod` to an `hdl21.ExternalModule` """
        # Check out cache for whether a module by the same name has been imported
        key = (pmod.name.domain, pmod.name.name)
        if key in self.ext_modules:
            conflict = self.ext_modules[key]
            raise RuntimeError(
                f"Cannot import conflicting definitions of {pmod} and {conflict}"
            )
        # Create the ExternalModule
        emod = ExternalModule(
            name=pmod.name.name,
            domain=pmod.name.domain,
            desc=pmod.desc,
            port_list=self.import_ports(pmod.ports),
            paramtype=object,  # FIXME: should these be stored in the serialization schema?
        )
        # Give it a (non-initializer) value for its `importpath`
        emod.importpath = [pmod.name.domain]
        # Cache and return it
        self.ext_modules[key] = emod
        return emod

    def import_module(self, pmod: protodefs.Module) -> Module:
        """ Convert Proto-Module `pmod` to an `hdl21.Module` """
        if pmod.name in self.modules:
            raise RuntimeError(f"Proto Import Error: Redefined Module {pmod.name}")

        # Create the Module
        module = Module()
        # Get or create its namespace, and set its name
        path = pmod.name.split(".")  # Path-parts are dot-separated
        ns = self.get_namespace(path[:-1])
        # Save the import-path and name
        module._importpath = path[:-1]
        module.name = path[-1]

        # Import its ports
        [module.add(port) for port in self.import_ports(pmod.ports)]

        # Import its signals
        for psig in pmod.signals:
            sig = Signal(name=psig.name, width=psig.width)
            module.add(sig)

        # Lap through instances, connecting them, creating internal Signals if necessary (bleh)
        for pinst in pmod.instances:
            inst = self.import_instance(pinst)
            module.add(inst)

            # Make the instance's connections
            for pname, pconn in pinst.connections.items():
                if pname not in inst._resolved.ports:
                    raise RuntimeError(
                        f"Invalid Port {pname} on {inst} in Module {module.name}"
                    )
                # Import the Signal-object
                sig = self.import_connection(pconn, module)
                # And connect it to the Instance
                setattr(inst, pname, sig)

        # Add the Module to our cache and return-namespace, and return it
        self.modules[pmod.name] = module
        setattr(ns, module.name, module)
        return module

    def import_connection(
        self, pconn: protodefs.Connection, module: Module
    ) -> Union[Signal, Slice, Concat]:
        """ Import a Proto-defined `Connection` into a Signal, Slice, or Concatenation """
        # Connections are a proto `oneof` union; figure out which to import
        stype = pconn.WhichOneof("stype")
        # Concatenations are more complicated and need their own method
        if stype == "concat":
            return self.import_concat(pconn.concat, module)
        # For signals & slices, first sort out the signal-name, so we can grab the object from `module.namespace`
        if stype == "sig":
            sname = pconn.sig.name
        elif stype == "slice":
            sname = pconn.slice.signal
        else:
            raise ValueError(f"Invalid Connection Type: {pconn}")
        # Grab this Signal, if it exists
        sig = module.namespace.get(sname, None)
        if sig is None:
            # This block has held, at some points in code-history,
            # the SPICE-style "create nets from thin air" behavior.
            # That's outta here; undeclared signals produce errors instead.
            raise RuntimeError(f"Invalid Signal {sname} in Module {module.name}")
        # Now chop this up if it's a Slice
        if stype == "slice":
            sig = Slice(signal=sig, top=pconn.slice.top, bot=pconn.slice.bot, step=1)
        return sig

    def import_concat(self, pconc: protodefs.Concat, module: Module) -> Concat:
        """ Import a (potentially nested) Concatenation """
        parts = []
        for ppart in pconc.parts:
            part = self.import_connection(ppart, module)
            parts.append(part)
        return Concat(*parts)

    def import_instance(self, pinst: protodefs.Instance) -> Instance:
        """ Convert Proto-Instance `pinst` to an `hdl21.Instance`. 
        Requires an available Module-definition to be referenced. 
        Connections are *not* performed inside this method. """

        # Also a small piece of proof that Google hates Python.
        ref = pinst.module
        if ref.WhichOneof("to") == "local":
            # Internally-defined Module
            module = self.modules.get(ref.local, None)
            if module is None:
                raise RuntimeError(f"Invalid undefined Module {ref.local} ")
            if len(pinst.parameters):
                raise RuntimeError(
                    f"Invalid Instance {pinst} with of Module {module} - does not accept Parameters"
                )
            target = module

        elif ref.WhichOneof("to") == "external":  # Defined outside package
            # First check the priviledged/ internally-defined domains
            if ref.external.domain in ("hdl21.primitives", "hdl21.ideal"):
                # Retrieve the Primitive from `hdl21.primitives`
                prim = self.import_primitive(ref.external)
                # Import all of its instance parameters, and convert them to its param-type
                pdict = self.import_parameters(pinst.parameters)
                params = prim.Params(**pdict)
                # Call the Primitive with its parameters, creating a PrimitiveCall Instance-target
                target = prim(params)

            else:  # External Module
                key = (ref.external.domain, ref.external.name)
                emod = self.ext_modules.get(key, None)
                if emod is None:
                    raise RuntimeError(
                        f"Invalid Instance of undefined External Module {key}"
                    )
                # Import all of its instance parameters to a dict
                pdict = self.import_parameters(pinst.parameters)
                # And call it with the parameters
                target = emod(pdict)

        return Instance(name=pinst.name, of=target)

    @classmethod
    def import_primitive(cls, pref: protodefs.QualifiedName) -> Primitive:
        if pref.domain not in ["hdl21.primitives", "hdl21.ideal"]:
            raise ValueError
        prim = getattr(primitives, pref.name, None)
        if not isinstance(prim, Primitive):
            raise RuntimeError(
                f"Attempt to import invalid `hdl21.primitive` {pref.external.name}"
            )
        return prim

    @classmethod
    def import_parameters(cls, pparams: Dict[str, protodefs.Parameter]) -> dict:
        pdict = {}
        for pname, pparam in pparams.items():
            pdict[pname] = cls.import_parameter(pparam)
        return pdict

    @classmethod
    def import_parameter(cls, pparam: protodefs.Parameter) -> Any:
        ptype = pparam.WhichOneof("value")
        if ptype == "integer":
            return int(pparam.integer)
        if ptype == "double":
            return float(pparam.double)
        if ptype == "string":
            return str(pparam.string)
        raise ValueError

    def import_ports(self, pports: List[protodefs.Port]) -> List[Signal]:
        # Import a list of proto-ports
        ports = []
        for pport in pports:
            dir_ = self.import_port_dir(pport)
            ports.append(
                Port(name=pport.signal.name, width=pport.signal.width, direction=dir_,)
            )
        return ports

    def import_port_dir(self, pport: protodefs.Port) -> PortDir:
        # Convert between Port-Direction Enumerations
        if pport.direction == protodefs.Port.Direction.INPUT:
            return PortDir.INPUT
        if pport.direction == protodefs.Port.Direction.OUTPUT:
            return PortDir.OUTPUT
        if pport.direction == protodefs.Port.Direction.INOUT:
            return PortDir.INOUT
        if pport.direction == protodefs.Port.Direction.NONE:
            return PortDir.NONE
        raise ValueError



""" 
# Hdl21 Built-In Generators Library 
"""

import copy
from dataclasses import asdict, replace
from typing import Optional, Tuple, Union

from . import primitives
from .primitives import MosType, MosVth
from .generator import generator
from .module import Module
from .params import paramclass, Param
from .signal import Signal
from .instantiable import Instantiable


@paramclass
class MosParams:
    """ Mos Series-Stack Generator Parameters """

    w = Param(dtype=Optional[int], desc="Width in resolution units", default=None)
    l = Param(dtype=Optional[int], desc="Length in resolution units", default=None)
    nser = Param(dtype=int, desc="Number of series fingers", default=1)
    npar = Param(dtype=int, desc="Number of parallel fingers", default=1)
    tp = Param(dtype=MosType, desc="MosType (PMOS/NMOS)", default=MosType.NMOS)
    vth = Param(dtype=MosVth, desc="Threshold voltage specifier", default=MosVth.STD)

    def __post_init_post_parse__(self):
        """ Value Checks """
        if self.w <= 0:
            raise ValueError(f"MosParams with invalid width {self.w}")
        if self.l <= 0:
            raise ValueError(f"MosParams with invalid length {self.l}")
        if self.npar <= 0:
            raise ValueError(
                f"MosParams with invalid number parallel fingers {self.npar}"
            )
        if self.nser <= 0:
            raise ValueError(
                f"MosParams with invalid number series fingers {self.nser}"
            )


@generator
def Mos(params: MosParams) -> Module:
    """ Mos Series-Stack Generator 
    Generates a `Module` including `nser` identical series instances of unit-Mos `primitives.Mos`.  
    Unit-Mos gate and bulk ports are connected in parallel. """

    # Extract the number of series fingers
    nser = params.nser
    # Extract the remaining params for the unit transistor
    unit_params = asdict(params)
    unit_params.pop("nser")
    unit_params = primitives.Mos.Params(**unit_params)
    # Create the Primitive unit-cell
    unit_xtor = primitives.Mos(unit_params)

    # Initialize our stack-module
    m = Module()
    # Copy the unit-cell ports
    for p in primitives.Mos.port_list:
        m.add(copy.copy(p))

    # Add instances, starting at the source-side
    inst = m.add(name="unit0", val=unit_xtor(s=m.s, g=m.g, b=m.b))
    for iser in range(1, nser):
        prev_inst = inst
        inst = m.add(name=f"unit{iser}", val=unit_xtor(s=prev_inst.d, g=m.g, b=m.b))
    # Finally connect the drain to the last instance
    inst.d = m.d
    # And return the module
    return m


@generator
def Nmos(params: MosParams) -> Module:
    """ Nmos Generator. A thin wrapper around `hdl21.generators.Mos` """
    return Mos(replace(params, tp=MosType.NMOS))


@generator
def Pmos(params: MosParams) -> Module:
    """ Pmos Constructor. A thin wrapper around `hdl21.generators.Mos` """
    return Mos(replace(params, tp=MosType.PMOS))


@paramclass
class SeriesParParams:
    """ Series-Parallel Generator Parameters """

    # Required
    unit = Param(dtype=Instantiable, desc="Unit cell")
    series_conns = Param(
        dtype=Tuple[Union[Signal, str], Union[Signal, str]],
        desc="Ports or port-names of `unit` to be connected in series",
    )
    # Optional
    nser = Param(dtype=int, desc="Number of series instances", default=1)
    npar = Param(dtype=int, desc="Number of parallel stacks", default=1)


@generator
def SeriesPar(params: SeriesParParams) -> Module:
    """ Series-Parallel Generator 
    Arrays `params.npar` copies of `params.nser` series-stacked Instances of unit-cell `params.unit`. 
    The generated `Module` includes the same ports as `unit`. 
    The two series-connected ports of `unit` are specified by parameter two-tuple `series_conns`. 
    All other ports of `unit` are wired in parallel, and exposed as ports of the generated `Module`. """

    unit = params.unit

    # Initialize our stack-module
    m = Module()
    # Copy the unit-cell ports
    for p in unit.ports.values():
        m.add(copy.copy(p))

    # Check for validity of the series-ports
    if isinstance(params.series_conns[0], str):
        ser0 = m.ports.get(params.series_conns[0], None)
    elif isinstance(params.series_conns[0], Signal):
        ser0 = params.series_conns[0]
    else:  # Unreachable
        raise TypeError
    if isinstance(params.series_conns[1], str):
        ser1 = m.ports.get(params.series_conns[1], None)
    elif isinstance(params.series_conns[1], Signal):
        ser1 = params.series_conns[1]
    else:  # Unreachable
        raise TypeError
    if ser0 is None or ser1 is None:
        raise ValueError(f"SeriesPar: unit does not have ports {params.series_conns}")

    # Extract all the parallel-connected ports
    par_conns = {
        port.name: port
        for port in unit.ports.values()
        if port.name not in params.series_conns
    }

    for ipar in range(params.npar):
        # Add instances, starting at the `series_conns[0]`-side
        inst = unit(**par_conns).connect(ser0.name, ser0)
        inst = m.add(name=f"unit_{ipar}_0", val=inst)
        for iser in range(1, params.nser):
            prev_inst = inst
            inst = unit(**par_conns)
            inst.connect(ser0.name, prev_inst.port_ref(ser1.name))
            inst = m.add(name=f"unit_{ipar}_{iser}", val=inst)
        # Finally connect the last series-port to the last instance
        inst.connect(ser1.name, ser1)
    # And return the module
    return m


def Wrapper(m: Module) -> Module:
    """ 
    # Module Wrapper Creator

    Adds a `Module` hierarchy layer around argument-`Module` `m`. 
    Creates an `Instance` of `m` and clones its ports, connecting each. 
    
    Note: `Wrapper` is generally aways more helpful when the returned `Module` is modified after the fact. 
    Hence while `Wrapper` is a *function that returns a `Module`*, it is *not* an `hdl21.Generator`, 
    which cache and unique-name their results. 
    Callers of `Wrapper` are therefore responsible for considerations such as unique naming. 
    """

    # FIXME: find this function a home with a less-confusing name!

    # Initialize our wrapper-module
    wrapper = Module(name=f"{m.name}Wrapper")

    # Copy the unit-cell ports
    for p in m.io.values():
        wrapper.add(copy.copy(p))

    # Create a connections-dict mirroring them
    conns = {port.name: port for port in wrapper.io.values()}

    # Create the inner instance
    wrapper.add(name="inner", val=m(**conns))

    # And return the wrapper
    return wrapper



"""
# hdl21 Signals and Ports 

Signals are Hdl21's base-level unit of hardware connectivity. 

Each `Signal` is analogous in content to a *bus* or (single-dimensional) *array* in many legacy HDLs. 
(Most similar to Verilog's packed single-dimensional arrays.) 
The `Signal.width` field indicates the bit-width of said bus. 
It defaults to one for scalar Signals. 
Widths of zero or less generate errors, both at construction-time and later. 

Hdl21 `Signals` are *untyped*. 
They represent a connection, not the type or value of data it carries. 
In this sense they are more similar to analog-style environments than to most legacy HDLs. 

`Ports`, `Inputs`, `Outputs`, and `Inouts` are not dedicated Hdl21 types, 
but thin convenience function-wrappers around `Signal`. 
Each `Signal` includes enumerated fields for its visibility (internal vs port) 
and direction. For internal `Signals`, the `direction` field is globally expected to be ignored. 

"""

from typing import Callable, Optional, Any, List, Union
from enum import Enum
from dataclasses import field
from pydantic.dataclasses import dataclass

# Local imports
from .connect import connectable, is_connectable


def _slice_(*, parent: Union["Signal", "Slice", "Concat"], key: Any) -> "Slice":
    """ Square-bracket slicing into Signals, Concatenations, and Slices. 
    Assuming valid inputs, returns a signal-`Slice`. 

    FIXME: this commentary reflects what *should* happen, not totally what *is* yet: 

    Signal slices are indexed "Python style", in the senses that: 
    * Negative indices are supported, and count from the "end" of the Signal.
    * Slice-ranges such as `sig[0:2]` are supported, and *inclusive* of the start, while *exclusive* of the end index. 
    * Negative-range slices such as `sig[2:0:-1]`, again *inclusive* of the start, *exclusive* of the end index, and *reversed*.
    Popular HDLs commonly use different signal-indexing conventions. 
    Hdl21's own primary exchange format (in ProtoBuf) does as well, 
    eschewing adopting inclusive-endpoints and eschewing negative-indexing.
    """

    if isinstance(key, int):
        if key >= parent.width:
            raise ValueError(f"Out-of-bounds index {key} into {parent}")
        if key < 0:
            key = key + parent.width
        return Slice(signal=parent, top=key, bot=key, step=1)

    if isinstance(key, slice):
        # Note these `slice` attributes are descriptor-things, and they get weird, fast.
        # Extracting their three key fields the most-hardest way via `__getattribute__` seems to work cleanest.
        start = slice.__getattribute__(key, "start")
        stop = slice.__getattribute__(key, "stop")
        step = slice.__getattribute__(key, "step")
        start = 0 if start is None else start
        stop = parent.width - 1 if stop is None else stop
        step = 1 if step is None else step
        # Wrap around the start and stop to allow for negative slicing
        start = start if start >= 0 else parent.width + start
        stop = stop if stop >= 0 else parent.width + stop - 1
        if step == 0:
            raise ValueError(f"slice step cannot be zero")
        elif step < 0:
            # Align bot with the step
            top = start
            bot = start + (-(start - stop) // step) * step
        else:
            # Align top with the step
            top = start + -(-(stop - start) // step) * step
            bot = start
        if top > parent.width:
            raise ValueError(f"Out-of-bounds index {top} into {parent}")
        if bot > parent.width:
            raise ValueError(f"Out-of-bounds index {bot} into {parent}")
        if bot > top:
            raise ValueError(f"Invalid slice ({bot} > {top}) {key} into {parent}")
        return Slice(signal=parent, top=top, bot=bot, step=step)

    raise TypeError(f"Invalid slice-type {key} into {parent}")


def slices(cls: type) -> type:
    """ Decorator to add the 'square-bracket indexing produces `Slice`s' behavior. """

    if getattr(cls, "__getitem__", None) is not None:
        msg = f"Internal hdl21 Error: invavlid `slices`-decoration of {cls} with existing __getitem__"
        raise RuntimeError(msg)

    def __getitem__(self, key: Any) -> "Slice":
        return _slice_(parent=self, key=key)

    # Add the new behavior to the class
    cls.__getitem__ = __getitem__
    cls.__getitem__.__doc__ = _slice_.__doc__
    # And don't forget to return that class!
    return cls


class PortDir(Enum):
    """ Port-Direction Enumeration """

    INPUT = 0
    OUTPUT = 1
    INOUT = 2
    NONE = 3


class Visibility(Enum):
    """ Port-Visibility Enumeration """

    INTERNAL = 0  # Internal, Module-private Signal
    PORT = 1  # Exposed as a Port


@slices
@connectable
@dataclass
class Signal:
    """
    # hdl21 Signal
    The base-level unit of hardware connectivity
    """

    name: Optional[str] = None
    width: int = 1
    vis: Visibility = Visibility.INTERNAL
    direction: PortDir = PortDir.NONE
    desc: Optional[str] = None  # Description
    src: Optional[Enum] = field(repr=False, default=None)
    dest: Optional[Enum] = field(repr=False, default=None)

    def __post_init_post_parse__(self):
        if self.width < 1:
            raise ValueError


def Signals(num: int, **kwargs) -> List[Signal]:
    """ 
    Create `num` new Signals.  
    Typical usage: 
    ```python
    @h.module 
    class UsesSignals:
        bias, fold, mirror = h.Signals(3)
    ```
    Note the `num` value is required to support the tuple-destructuring use-case shown above. 
    """
    return _plural(fn=Signal, num=num, **kwargs)


def Input(**kwargs) -> Signal:
    """ Input Port Constructor. Thin wrapper around `hdl21.Signal` """
    return Signal(vis=Visibility.PORT, direction=PortDir.INPUT, **kwargs)


def Inputs(num: int, **kwargs) -> List[Signal]:
    """ 
    Create `num` new Input Ports.  
    Typical usage: 
    ```python
    @h.module 
    class UsesInputs:
        a, b, c, VDD, VSS = h.Inputs(5)
    ```
    Note the `num` value is required to support the tuple-destructuring use-case shown above. 
    """
    return _plural(fn=Input, num=num, **kwargs)


def Output(**kwargs) -> Signal:
    """ Output Port Constructor. Thin wrapper around `hdl21.Signal` """
    return Signal(vis=Visibility.PORT, direction=PortDir.OUTPUT, **kwargs)


def Outputs(num: int, **kwargs) -> List[Signal]:
    """ 
    Create `num` new Output Ports.  
    Typical usage: 
    ```python
    @h.module 
    class UsesOutputs:
        tdo, tms, tck = h.Outputs(3)
    ```
    Note the `num` value is required to support the tuple-destructuring use-case shown above. 
    """
    return _plural(fn=Output, num=num, **kwargs)


def Inout(**kwargs) -> Signal:
    """ Inout Port Constructor. Thin wrapper around `hdl21.Signal` """
    return Signal(vis=Visibility.PORT, direction=PortDir.INOUT, **kwargs)


def Inouts(num: int, **kwargs) -> List[Signal]:
    """ 
    Create `num` new Inout Ports.  
    Typical usage: 
    ```python
    @h.module 
    class UsesInouts:
        gpio1, gpio2 = h.Inouts(2)
    ```
    Note the `num` value is required to support the tuple-destructuring use-case shown above. 
    """
    return _plural(fn=Inout, num=num, **kwargs)


def Port(direction=PortDir.NONE, **kwargs) -> Signal:
    """ Port Constructor. Thin wrapper around `hdl21.Signal`.
    The `direction` argument sets the Port's direction,
    and defaults to the unknown direction `PortDir.NONE`. """
    return Signal(direction=direction, vis=Visibility.PORT, **kwargs)


def Ports(num: int, **kwargs) -> List[Signal]:
    """ 
    Create `num` new Ports.  
    Typical usage: 
    ```python
    @h.module 
    class UsesPorts:
        inp, out = h.Ports(2)
    ```
    Note the `num` value is required to support the tuple-destructuring use-case shown above. 
    """
    return _plural(fn=Port, num=num, **kwargs)


def _plural(*, fn: Callable, num: int, **kwargs) -> List[Signal]:
    """ Internal helper method for creating `num` identical `Signal` objects via callable `fn`. """
    rv = list()
    for _ in range(num):
        rv.append(fn(**kwargs))
    return rv


@slices
@connectable
class Concat:
    """ Signal Concatenation
    Uses *Python-convention* ordering, in which "LSBs", i.e. index 0, are specified first. """

    def __init__(self, *parts):
        for p in parts:
            if not is_connectable(p):
                raise TypeError(f"Signal-concatenating unconnectable object {p}")
        self.parts = tuple(parts)

    @property
    def width(self):
        return sum([s.width for s in self.parts])

    def __eq__(self, _other) -> bool:
        # Concat-equality can be implemented, but has plenty of edge cases
        # not yet worked through. For example slicing and re-concatenation:
        # ```python
        # s = h.Signal(width=2)      # Create a 2-bit signal
        # s == h.Concat(s[0], s[1])  # Slice and re-concatenate it
        # # Should that be True of False?
        # ```
        raise NotImplementedError

    def __repr__(self):
        return f"Concat(width={self.width}, parts={len(self.parts)})"


@slices
@connectable
@dataclass
class Slice:
    """ Signal Slice, comprising a subset of its width """

    signal: Union[Signal, Concat, "Slice"]  # Parent Signal
    top: int  # Top index (inclusive)
    bot: int  # Bottom index (inclusive)
    step: int  # Index step size

    @property
    def width(self) -> int:
        """ Slice width """
        if self.step != 1:
            raise NotImplementedError  # FIXME!
        return 1 + self.top - self.bot

    def __eq__(self, other: "Slice") -> bool:
        """ Slice equality requires *identity* between parent Signals """
        if not isinstance(other, Slice):
            return NotImplemented
        return (
            self.signal is other.signal
            and self.top == other.top
            and self.bot == other.bot
            and self.step == other.step
        )


Slice.__pydantic_model__.update_forward_refs()


@connectable
@dataclass
class NoConn:
    """
    # No-Connect 
    
    Special placeholder connectable-object which indicates "unconnected" Ports,
    typically unconnected outputs.
    
    An optional `name` field allows guidance for external netlisting, 
    for cases in which consistent naming is desirable (e.g. for waveform probing). 
    """

    name: Optional[str] = None



"""
# Hdl21 Primitive Modules

Primitives are leaf-level Modules typically defined not by users, 
but by simulation tools or device fabricators. 
Prominent examples include MOS transistors, diodes, resistors, and capacitors. 

Primitives divide in two classes, `physical` and `ideal`, 
indicated by their `primtype` attribute. 
`PrimitiveType.IDEAL` primitives specify circuit-theoretic ideal elements 
e.g. resistors, capacitors, inductors, and notably aphysical elements 
such as ideal voltage and current sources. 
`PrimitiveType.PHYSICAL` primitives in contrast specify abstract versions 
of ultimately physically-realizable elements such as transistors and diodes. 
These elements typically require some external translation, e.g. by a process-technology 
library, to execute in simulations or to be realized in hardware. 

Many element-types (particularly passives) come in both `ideal` and `physical` flavors, 
as typical process-technologies include physical passives, but with far different 
parameterization than ideal passives. For example resistors are commonly specified 
in physical length and width. Capacitors are similarly specified in physical terms, 
often adding metal layers or other physical features. The component-value (R,C,L, etc.) 
for these physically-specified cells is commonly suggestive or optional. 

| Physical           | Ideal          | Alias(es)         | 
| ------------------ | -------------- | ----------------- | 
| PhysicalResistor   | IdealResistor  | R, Res, Resistor  | 
| PhysicalInductor   | IdealInductor  | L, Ind, Inductor  | 
| PhysicalCapacitor  | IdealCapacitor | C, Cap, Capacitor | 
| PhysicalShort      | IdealShort     | Short             | 
|                    | VoltageSource  | Vsrc, V           | 
|                    | CurrentSource  | Isrc, I           | 
| Mos                |                |                   | 
| Diode              |                | D                 | 

"""

from pydantic.dataclasses import dataclass
from dataclasses import replace
from enum import Enum
from typing import Optional, Any, List, Type, Union

# Local imports
from .params import paramclass, Param, isparamclass, HasNoParams
from .signal import Port, Signal, Visibility
from .instance import calls_instantiate


class PrimitiveType(Enum):
    """ Enumerated Primitive-Types """

    IDEAL = "IDEAL"
    PHYSICAL = "PHYSICAL"


@dataclass
class Primitive:
    """# hdl21 Primitive Component

    Primitives are leaf-level Modules typically defined not by users,
    but by simulation tools or device fabricators.
    Prominent examples include MOS transistors, diodes, resistors, and capacitors.
    """

    name: str
    desc: str
    port_list: List[Signal]
    paramtype: Type
    primtype: PrimitiveType

    def __post_init_post_parse__(self):
        """After type-checking, do plenty more checks on values"""
        if not isparamclass(self.paramtype):
            raise TypeError(
                f"Invalid Primitive param-type {self.paramtype} for {self.name}, must be an `hdl21.paramclass`"
            )
        for p in self.port_list:
            if not p.name:
                raise ValueError(f"Unnamed Primitive Port {p} for {self.name}")
            if p.vis != Visibility.PORT:
                raise ValueError(
                    f"Invalid Primitive Port {p.name} on {self.name}; must have PORT visibility"
                )

    def __call__(self, params: Any) -> "PrimitiveCall":
        return PrimitiveCall(prim=self, params=params)

    @property
    def Params(self) -> Type:
        return self.paramtype

    @property
    def ports(self) -> dict:
        return {p.name: p for p in self.port_list}


@calls_instantiate
@dataclass
class PrimitiveCall:
    """Primitive Call
    A combination of a Primitive and its Parameter-values,
    typically generated by calling the Primitive."""

    prim: Primitive
    params: Any

    def __post_init_post_parse__(self):
        # Type-validate our parameters
        if not isinstance(self.params, self.prim.paramtype):
            raise TypeError(
                f"Invalid parameters {self.params} for Primitive {self.prim}. Must be {self.prim.paramtype}"
            )

    @property
    def ports(self) -> dict:
        return self.prim.ports


class MosType(Enum):
    """NMOS/PMOS Type Enumeration"""

    NMOS = "NMOS"
    PMOS = "PMOS"


class MosVth(Enum):
    """MOS Threshold Enumeration"""

    STD = "STD"
    LOW = "LOW"
    HIGH = "HIGH"


@paramclass
class MosParams:
    """MOS Transistor Parameters"""

    w = Param(dtype=Optional[int], desc="Width in resolution units", default=None)
    l = Param(dtype=Optional[int], desc="Length in resolution units", default=None)
    npar = Param(dtype=int, desc="Number of parallel fingers", default=1)
    tp = Param(dtype=MosType, desc="MosType (PMOS/NMOS)", default=MosType.NMOS)
    vth = Param(dtype=MosVth, desc="Threshold voltage specifier", default=MosVth.STD)

    def __post_init_post_parse__(self):
        """ Value Checks """
        if self.w <= 0:
            raise ValueError(f"MosParams with invalid width {self.w}")
        if self.l <= 0:
            raise ValueError(f"MosParams with invalid length {self.l}")
        if self.npar <= 0:
            raise ValueError(
                f"MosParams with invalid number parallel fingers {self.npar}"
            )
        if self.nser <= 0:
            raise ValueError(
                f"MosParams with invalid number series fingers {self.nser}"
            )


MosPorts = [Port(name="d"), Port(name="g"), Port(name="s"), Port(name="b")]

Mos = Primitive(
    name="Mos",
    desc="Mos Transistor",
    port_list=MosPorts,
    paramtype=MosParams,
    primtype=PrimitiveType.PHYSICAL,
)


def Nmos(params: MosParams) -> Primitive:
    """ Nmos Constructor. A thin wrapper around `hdl21.primitives.Mos` """
    return Mos(replace(params, tp=MosType.NMOS))


def Pmos(params: MosParams) -> Primitive:
    """ Pmos Constructor. A thin wrapper around `hdl21.primitives.Mos` """
    return Mos(replace(params, tp=MosType.PMOS))


@paramclass
class DiodeParams:
    w = Param(dtype=Optional[int], desc="Width in resolution units", default=None)
    l = Param(dtype=Optional[int], desc="Length in resolution units", default=None)
    # FIXME: will likely want a similar type-switch, at least eventually
    # tp = Param(dtype=Tbd!, desc="Diode type specifier")


Diode = Primitive(
    name="Diode",
    desc="Diode",
    port_list=[Port(name="p"), Port(name="n")],
    paramtype=DiodeParams,
    primtype=PrimitiveType.PHYSICAL,
)


# Common alias(es)
D = Diode


@paramclass
class ResistorParams:
    r = Param(dtype=float, desc="Resistance (ohms)")


IdealResistor = Primitive(
    name="IdealResistor",
    desc="Ideal Resistor",
    port_list=[Port(name="p"), Port(name="n")],
    paramtype=ResistorParams,
    primtype=PrimitiveType.IDEAL,
)


# Common aliases
R = Res = Resistor = IdealResistor


@paramclass
class PhysicalResistorParams:
    r = Param(dtype=float, desc="Resistance (ohms)")


PhysicalResistor = Primitive(
    name="PhysicalResistor",
    desc="Physical Resistor",
    port_list=[Port(name="p"), Port(name="n")],
    paramtype=ResistorParams,
    primtype=PrimitiveType.PHYSICAL,
)


@paramclass
class IdealCapacitorParams:
    c = Param(dtype=float, desc="Capacitance (F)")


IdealCapacitor = Primitive(
    name="IdealCapacitor",
    desc="Ideal Capacitor",
    port_list=[Port(name="p"), Port(name="n")],
    paramtype=IdealCapacitorParams,
    primtype=PrimitiveType.IDEAL,
)


# Common aliases
C = Cap = Capacitor = IdealCapacitor


@paramclass
class PhysicalCapacitorParams:
    c = Param(dtype=float, desc="Capacitance (F)")


PhysicalCapacitor = Primitive(
    name="PhysicalCapacitor",
    desc="Physical Capacitor",
    port_list=[Port(name="p"), Port(name="n")],
    paramtype=PhysicalCapacitorParams,
    primtype=PrimitiveType.PHYSICAL,
)


@paramclass
class IdealInductorParams:
    l = Param(dtype=float, desc="Inductance (H)")


IdealInductor = Primitive(
    name="IdealInductor",
    desc="Ideal Inductor",
    port_list=[Port(name="p"), Port(name="n")],
    paramtype=IdealInductorParams,
    primtype=PrimitiveType.IDEAL,
)


# Common alias(es)
L = Inductor = IdealInductor


@paramclass
class PhysicalInductorParams:
    l = Param(dtype=float, desc="Inductance (H)")


PhysicalInductor = Primitive(
    name="PhysicalInductor",
    desc="Physical Inductor",
    port_list=[Port(name="p"), Port(name="n")],
    paramtype=PhysicalInductorParams,
    primtype=PrimitiveType.PHYSICAL,
)


@paramclass
class PhysicalShortParams:
    layer = Param(dtype=Optional[int], desc="Metal layer", default=None)
    w = Param(dtype=Optional[int], desc="Width in resolution units", default=None)
    l = Param(dtype=Optional[int], desc="Length in resolution units", default=None)


PhysicalShort = Primitive(
    name="PhysicalShort",
    desc="Short-Circuit/ Net-Tie",
    port_list=[Port(name="p"), Port(name="n")],
    paramtype=PhysicalShortParams,
    primtype=PrimitiveType.PHYSICAL,
)

IdealShort = Primitive(
    name="IdealShort",
    desc="Short-Circuit/ Net-Tie",
    port_list=[Port(name="p"), Port(name="n")],
    paramtype=HasNoParams,
    primtype=PrimitiveType.IDEAL,
)

Short = IdealShort


# Type alias for many scalar parameters
ScalarParam = Union[int, float, str]
ScalarOption = Optional[ScalarParam]


@paramclass
class VoltageSourceParams:
    dc = Param(dtype=ScalarOption, default=0, desc="DC Value (V)")
    delay = Param(dtype=ScalarOption, default=None, desc="Time Delay (s)")

    # Pulse source parameters
    v0 = Param(dtype=ScalarOption, default=None, desc="Zero Value (V)")
    v1 = Param(dtype=ScalarOption, default=None, desc="One Value (V)")
    period = Param(dtype=ScalarOption, default=None, desc="Period (s)")
    rise = Param(dtype=ScalarOption, default=None, desc="Rise time (s)")
    fall = Param(dtype=ScalarOption, default=None, desc="Fall time (s)")
    width = Param(dtype=ScalarOption, default=None, desc="Pulse width (s)")


VoltageSource = Primitive(
    name="VoltageSource",
    desc="Ideal Voltage Source",
    port_list=[Port(name="p"), Port(name="n")],
    paramtype=VoltageSourceParams,
    primtype=PrimitiveType.IDEAL,
)

V = Vsrc = VoltageSource


@paramclass
class CurrentSourceParams:
    dc = Param(dtype=ScalarOption, default=0, desc="DC Value (A)")


CurrentSource = Primitive(
    name="CurrentSource",
    desc="Ideal Current Source",
    port_list=[Port(name="p"), Port(name="n")],
    paramtype=CurrentSourceParams,
    primtype=PrimitiveType.IDEAL,
)

I = Isrc = CurrentSource




""" 
# Hdl21 Hardware Description Library 
"""

__version__ = "0.1.1"


# Before any real importing, ensure we can instantiate non-pydantic types in type-checked dataclasses.
# This `Config` seems to be shared for *all* pydantic types, even when not applied to `BaseModel`.
from pydantic import BaseModel

BaseModel.Config.arbitrary_types_allowed = True

from .params import *
from .instance import *
from .signal import *
from .module import *
from .generator import *
from .generators import *
from .primitives import *
from .interface import *
from .netlist import *
from .elab import *
from .walker import HierarchyWalker
from .instantiable import *

from . import proto
from .proto import to_proto, from_proto

